import CADETMatch.util as util
from CADETMatch.abstract.transform import AbstractTransform


class SetTransform(AbstractTransform):
    @property
    def name(self):
        return "set_value"

    @property
    def count(self):
        return 0

    @property
    def count_extended(self):
        return 0

    def transform(self):
        return []

    grad_transform = transform

    def untransform_inputorder(self, seq):
        return []

    def untransform(self, seq):
        return [], []

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        return None

    untransform_matrix_inputorder = untransform_matrix

    def setSimulation(self, sim, seq, experiment):
        if self.parameter.get("experiments", None) is None or experiment["name"] in self.parameter["experiments"]:
            locationFrom = self.parameter["locationFrom"]
            locationTo = self.parameter["locationTo"]

            try:
                compFrom = self.parameter["componentFrom"]
                boundFrom = self.parameter["boundFrom"]
                indexFrom = None
            except KeyError:
                indexFrom = self.parameter["indexFrom"]
                boundFrom = None
                compFrom = None
            valueFrom = self.getValue(sim, locationFrom, bound=boundFrom, comp=compFrom, index=indexFrom)

            try:
                compTo = self.parameter["componentTo"]
                boundTo = self.parameter["boundTo"]
                indexTo = None
            except KeyError:
                indexTo = self.parameter["indexTo"]
                boundTo = None
                compTo = None
            self.setValue(sim, valueFrom, locationTo, bound=boundTo, comp=compTo, index=indexTo)

        return [], []

    def getBounds(self):
        return None, None

    def getGradBounds(self):
        return None, None

    def getHeaders(self):
        return []

    def getHeadersActual(self):
        return self.getHeaders()

    def setBounds(self, parameter, lb, ub):
        return None


plugins = {"set_value": SetTransform}



import importlib
import importlib.util
import sys

from pathlib import Path

base = Path(__file__).parent


def load_plugin(path):
    module = ".".join(path.relative_to(base).parts).replace(".py", "")
    spec = importlib.util.spec_from_file_location(module, str(path))
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    sys.modules[foo.__name__] = foo
    return foo


def get_files(dir):
    return [path for path in dir.glob("*.py") if not path.name == "__init__.py"]


def get_plugins(directory):
    plugins = base / directory
    temp = {}
    for path in get_files(plugins):
        plug = load_plugin(path)
        if getattr(plug, "plugins", None) is not None:
            temp.update(plug.plugins)
        else:
            temp[plug.name] = plug
    return temp



import CADETMatch.util as util
import numpy
from CADETMatch.abstract.transform import AbstractTransform


class KeqTransform(AbstractTransform):
    @property
    def name(self):
        return "keq"

    @property
    def count(self):
        return 2

    @property
    def count_extended(self):
        return 3

    def transform(self):
        def trans_a(i):
            return numpy.log(i)

        def trans_b(i):
            return numpy.log(i)

        return [trans_a, trans_b]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        ka = numpy.exp(seq[0])
        keq = numpy.exp(seq[1])
        return [ka, keq]

    def untransform(self, seq):
        ka, keq = self.untransform_inputorder(seq)
        kd = ka/keq
        values = [ka, kd]
        headerValues = [ka, kd, keq]
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = self.untransform_matrix_inputorder(matrix)
        values[:, 1] = values[:, 0] / values[:, 1]
        return values

    def untransform_matrix_inputorder(self, matrix):
        values = numpy.exp(matrix)
        return values

    def setSimulation(self, sim, seq, experiment):
        values, headerValues = self.untransform(seq)

        if self.parameter.get("experiments", None) is None or experiment["name"] in self.parameter["experiments"]:
            location = self.parameter["location"]

            comp = self.parameter["component"]
            bound = self.parameter["bound"]

            unit = self.getUnit(location[0])
            boundOffset = util.getBoundOffset(sim.root.input.model[unit])

            position = boundOffset[comp] + bound
            sim[location[0].lower()][position] = values[0]
            sim[location[1].lower()][position] = values[1]

        return values, headerValues

    def getBounds(self):
        minKA = self.parameter["minKA"]
        maxKA = self.parameter["maxKA"]
        minKEQ = self.parameter["minKEQ"]
        maxKEQ = self.parameter["maxKEQ"]

        minValues = list(numpy.log([minKA, minKEQ]))
        maxValues = list(numpy.log([maxKA, maxKEQ]))

        return minValues, maxValues

    def getGradBounds(self):
        return [self.parameter["min"],], [self.parameter["max"],]

    def getHeaders(self):
        location = self.parameter["location"]
        nameKA = location[0].rsplit("/", 1)[-1]
        nameKD = location[1].rsplit("/", 1)[-1]
        bound = self.parameter["bound"]
        comp = self.parameter["component"]

        headers = []
        headers.append("%s Comp:%s Bound:%s" % (nameKA, comp, bound))
        headers.append("%s Comp:%s Bound:%s" % (nameKD, comp, bound))
        headers.append("%s/%s Comp:%s Bound:%s" % (nameKA, nameKD, comp, bound))
        return headers

    def getHeadersActual(self):
        location = self.parameter["location"]
        nameKA = location[0].rsplit("/", 1)[-1]
        nameKD = location[1].rsplit("/", 1)[-1]
        bound = self.parameter["bound"]
        comp = self.parameter["component"]

        headers = []
        headers.append("%s Comp:%s Bound:%s" % (nameKA, comp, bound))
        headers.append("%s/%s Comp:%s Bound:%s" % (nameKA, nameKD, comp, bound))
        return headers

    def setBounds(self, parameter, lb, ub):
        parameter["minKA"] = lb[0]
        parameter["maxKA"] = ub[0]
        parameter["minKEQ"] = lb[1]
        parameter["maxKEQ"] = ub[1]


class NormKeqTransform(KeqTransform):
    @property
    def name(self):
        return "norm_keq"

    def transform(self):
        minKA = numpy.log(self.parameter["minKA"])
        maxKA = numpy.log(self.parameter["maxKA"])
        minKEQ = numpy.log(self.parameter["minKEQ"])
        maxKEQ = numpy.log(self.parameter["maxKEQ"])

        def trans_a(i):
            return (numpy.log(i) - minKA) / (maxKA - minKA)

        def trans_b(i):
            return (numpy.log(i) - minKEQ) / (maxKEQ - minKEQ)

        return [trans_a, trans_b]

    grad_transform = transform


    def untransform_inputorder(self, seq):
        minKA = self.parameter["minKA"]
        maxKA = self.parameter["maxKA"]
        minKEQ = self.parameter["minKEQ"]
        maxKEQ = self.parameter["maxKEQ"]

        minValues = numpy.log([minKA, minKEQ])
        maxValues = numpy.log([maxKA, maxKEQ])

        values = numpy.array(seq)

        values = (maxValues - minValues) * values + minValues

        ka = numpy.exp(values[0])
        keq = numpy.exp(values[1])

        return [ka, keq]


    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = self.untransform_matrix_inputorder(matrix)
        values[:, 1] = values[:, 0] / values[:, 1]
        return values

    def untransform_matrix_inputorder(self, matrix):
        minKA = self.parameter["minKA"]
        maxKA = self.parameter["maxKA"]
        minKEQ = self.parameter["minKEQ"]
        maxKEQ = self.parameter["maxKEQ"]

        minValues = numpy.log([minKA, minKEQ])
        maxValues = numpy.log([maxKA, maxKEQ])

        values = (maxValues - minValues) * matrix + minValues

        values = numpy.exp(values)

        return values

    def getBounds(self):
        return [0.0, 0.0], [1.0, 1.0]

    def getGradBounds(self):
        return [self.parameter["min"],], [self.parameter["max"],]


plugins = {"keq": KeqTransform, "norm_keq": NormKeqTransform}



import CADETMatch.util as util
import sys
from addict import Dict
import numpy

name = "SSE"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = False
    settings.badScore = -sys.float_info.max
    settings.meta_mask = True
    settings.count = 1
    return settings


def run(sim_data, feature):
    "sum square error score, this score is NOT composable with other scores, use negative so score is maximized like other scores"
    sim_time_values, sim_data_values = util.get_times_values(sim_data["simulation"], feature)
    selected = feature["selected"]

    exp_time_values = feature["time"][selected]
    exp_data_values = feature["value"][selected]

    sse = util.sse(sim_data_values, exp_data_values)

    return ([-sse,], sse, len(sim_data_values), sim_time_values, sim_data_values, exp_data_values, [sse,])


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    temp["peak_max"] = max(selectedValues)
    return temp


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    temp = ["%s_SSE" % name]
    return temp



import numpy


def linear_coeff(x1, y1, x2, y2):
    "return paramters that fit y = a*x+b"
    a = (y2 - y1) / (x2 - x1)
    b = y2 - a * x2
    return a, b


def exponential_coeff(x1, y1, x2, y2):
    "return paramaters that fit y = b*exp(m*x)"
    b = (numpy.log(y2) - numpy.log(y1)) / (x2 - x1)
    a = y1 * numpy.exp(-b * x1)
    return a, b


def exponential(x, a, b):
    return a * numpy.exp(b * x)


def linear(x, a, b):
    return a * x + b



import random
import pickle
import CADETMatch.util as util
import CADETMatch.progress as progress
import numpy
import numpy as np
import scipy
from pathlib import Path
import time
import csv
import cadet

import emcee
import SALib.sample.sobol_sequence

import subprocess
import sys

import CADETMatch.evo as evo
import CADETMatch.cache as cache

import CADETMatch.pareto as pareto

import multiprocessing
import pandas
import array
import emcee.autocorr as autocorr
from sklearn.cluster import KMeans
import scipy.spatial

import CADETMatch.kde_generator as kde_generator

name = "MCMC"

from addict import Dict

import joblib

import CADETMatch.de as de
import CADETMatch.de_snooker as de_snooker
import CADETMatch.stretch as stretch

import jstyleson
import shutil

log2 = numpy.log(2)
min_acceptance = 0.05
acceptance_delta = 0.02

def log_previous(cadetValues, kde_previous, kde_previous_scaler):
    # find the right values to use
    col = len(kde_previous_scaler.scale_)
    values = cadetValues[-col:]
    values_shape = numpy.array(values).reshape(1, -1)
    values_scaler = kde_previous_scaler.transform(values_shape)
    score = kde_previous.score_samples(values_scaler)
    return score


def log_likelihood(individual, json_path):
    if json_path != cache.cache.json_path:
        cache.cache.setup_dir(json_path)
        util.setupLog(cache.cache.settings["resultsDirLog"], "main.log")
        cache.cache.setup(json_path, False)
        
    if "kde_previous" not in log_likelihood.__dict__:
        kde_previous, kde_previous_scaler = kde_generator.getKDEPrevious(cache.cache)
        log_likelihood.kde_previous = kde_previous
        log_likelihood.kde_previous_scaler = kde_previous_scaler

    if "kde" not in log_likelihood.__dict__:
        kde, kde_scaler = kde_generator.getKDE(cache.cache)
        log_likelihood.kde = kde
        log_likelihood.scaler = kde_scaler

    scores, csv_record, meta_score, results, individual = evo.fitness(individual, json_path)

    if results is None:
        multiprocessing.get_logger().info("log_likelihood results is None %s (%s)", individual, util.convert_individual_inputorder(individual, cache.cache))
        return -numpy.inf, scores, csv_record, meta_score, results, individual

    if results is not None and log_likelihood.kde_previous is not None:
        logPrevious = log_previous(individual, log_likelihood.kde_previous, log_likelihood.kde_previous_scaler)
    else:
        logPrevious = 0.0

    scores_shape = numpy.array(scores)
    scores_shape = scores_shape.reshape(1, -1)

    score_scaler = log_likelihood.scaler.transform(scores_shape)

    score_kde = log_likelihood.kde.score_samples(score_scaler)

    score = score_kde + log2 + logPrevious
    # *2 is from mirroring and we need to double the probability to get back to the normalized distribution

    return score, scores, csv_record, meta_score, results, individual


def log_posterior_vectorize(population, json_path, cache, halloffame, meta_hof, grad_hof, progress_hof, result_data, writer, csvfile):
    results = cache.toolbox.map(log_posterior, ((population[i], json_path) for i in range(len(population))))
    results = process(population, cache, halloffame, meta_hof, grad_hof, progress_hof, result_data, results, writer, csvfile)
    return results


def outside_bounds(x, cache):
    for i, lb, ub in zip(x, cache.MIN_VALUE, cache.MAX_VALUE):
        if i < lb or i > ub:
            return True
    return False

def log_posterior(x):
    theta, json_path = x
    if json_path != cache.cache.json_path:
        cache.cache.setup_dir(json_path)
        util.setupLog(cache.cache.settings["resultsDirLog"], "main.log")
        cache.cache.setup(json_path)

    if outside_bounds(theta, cache.cache):
        multiprocessing.get_logger().info("individual is outside of range %s (%s)", theta, util.convert_individual_inputorder(theta, cache.cache))
        return -numpy.inf, theta, cache.cache.WORST, [], cache.cache.WORST_META, None, theta

    ll, scores, csv_record, meta_score, results, individual = log_likelihood(theta, json_path)

    if results is None:
        multiprocessing.get_logger().info("log_posterior results is None %s (%s)", theta, util.convert_individual_inputorder(theta, cache.cache))
        return -numpy.inf, theta, cache.cache.WORST, [], cache.cache.WORST_META, None, individual
    else:
        return ll, theta, scores, csv_record, meta_score, results, individual


def addChain(*args):
    temp = [arg for arg in args if arg is not None]
    if len(temp) > 1:
        return numpy.concatenate(temp, axis=1)
    else:
        return numpy.array(temp[0])


def converged_bounds(chain, length, error_level):
    if chain.shape[1] < (2 * length):
        return False, None, None, None
    lb = []
    ub = []
    mid = []

    start = chain.shape[1] - length
    stop = chain.shape[1]
    for i in range(start, stop):
        temp_chain = chain[:, :i, :]
        temp_chain_shape = temp_chain.shape
        temp_chain_flat = temp_chain.reshape(temp_chain_shape[0] * temp_chain_shape[1], temp_chain_shape[2])
        lb_5, mid_50, ub_95 = numpy.percentile(temp_chain_flat, [5, 50, 95], 0)

        lb.append(lb_5)
        ub.append(ub_95)
        mid.append(mid_50)

    lb = numpy.array(lb)
    ub = numpy.array(ub)
    mid = numpy.array(mid)

    if numpy.all(numpy.std(lb, axis=0) < error_level) and numpy.all(numpy.std(ub, axis=0) < error_level):
        return True, numpy.mean(lb, axis=0), numpy.mean(mid, axis=0), numpy.mean(ub, axis=0)
    else:
        multiprocessing.get_logger().info(
            "bounds have not yet converged lb min: %s max: %s std: %s  ub min %s max: %s std: %s",
            numpy.array2string(numpy.min(lb, axis=0), precision=3, separator=","),
            numpy.array2string(numpy.max(lb, axis=0), precision=3, separator=","),
            numpy.array2string(numpy.std(lb, axis=0), precision=4, separator=","),
            numpy.array2string(numpy.min(ub, axis=0), precision=3, separator=","),
            numpy.array2string(numpy.max(ub, axis=0), precision=3, separator=","),
            numpy.array2string(numpy.std(ub, axis=0), precision=4, separator=","),
        )
        return False, None, None, None


def rescale(cache, lb, mid, ub, old_lb, old_ub, mcmc_store):
    "give a new lb and ub that will rescale so that the previous lb and ub takes up about 1/2 of the search width"
    new_size = len(lb)
    old_lb_slice = old_lb[:new_size]
    old_ub_slice = old_ub[:new_size]

    center = mid
    old_center = (old_ub + old_lb) / 2.0

    new_lb = lb - 2 * (center - lb)

    new_lb = numpy.max([new_lb, old_lb_slice], axis=0)

    new_ub = ub + 2 * (ub - center)

    new_ub = numpy.min([new_ub, old_ub_slice], axis=0)

    lb_trans = numpy.ones([3, len(old_lb)]) * old_lb
    ub_trans = numpy.ones([3, len(old_ub)]) * old_ub
    center_trans = numpy.ones([3, len(old_ub)]) * old_center

    lb_trans[1, :new_size] = lb
    lb_trans[2, :new_size] = new_lb

    ub_trans[1, :new_size] = ub
    ub_trans[2, :new_size] = new_ub

    center_trans[1, :new_size] = center
    center_trans[2, :new_size] = center

    lb_trans_conv = util.convert_population_inputorder(lb_trans, cache)
    center_trans_conv = util.convert_population_inputorder(center_trans, cache)
    ub_trans_conv = util.convert_population_inputorder(ub_trans, cache)

    mcmc_store.root.bounds_change.lb_trans = lb_trans
    mcmc_store.root.bounds_change.ub_trans = ub_trans
    mcmc_store.root.bounds_change.center_trans = center_trans
    mcmc_store.root.bounds_change.lb_trans_conv = lb_trans_conv
    mcmc_store.root.bounds_change.center_trans_conv = center_trans_conv
    mcmc_store.root.bounds_change.ub_trans_conv = ub_trans_conv

    multiprocessing.get_logger().info(
        """rescaling bounds (simulator space)  \nold_lb: %s \nold_center: %s \nold_ub: %s
    \nlb5: %s \ncenter: %s \nub5: %s
    \nnew_lb: %s \nnew_center: %s \nnew_ub: %s
    \nbounds (search space)
    \nold_lb: %s \nold_center: %s \nold_ub: %s
    \nlb5: %s \ncenter: %s \nub5: %s
    \nnew_lb: %s \nnew_center: %s \nnew_ub: %s""",
        numpy.array2string(lb_trans_conv[0], precision=3, separator=","),
        numpy.array2string(center_trans_conv[0], precision=3, separator=","),
        numpy.array2string(ub_trans_conv[0], precision=3, separator=","),
        numpy.array2string(lb_trans_conv[1], precision=3, separator=","),
        numpy.array2string(center_trans_conv[1], precision=3, separator=","),
        numpy.array2string(ub_trans_conv[1], precision=3, separator=","),
        numpy.array2string(lb_trans_conv[2], precision=3, separator=","),
        numpy.array2string(center_trans_conv[2], precision=3, separator=","),
        numpy.array2string(ub_trans_conv[2], precision=3, separator=","),
        numpy.array2string(lb_trans[0], precision=3, separator=","),
        numpy.array2string(center_trans[0], precision=3, separator=","),
        numpy.array2string(ub_trans[0], precision=3, separator=","),
        numpy.array2string(lb_trans[1], precision=3, separator=","),
        numpy.array2string(center_trans[1], precision=3, separator=","),
        numpy.array2string(ub_trans[1], precision=3, separator=","),
        numpy.array2string(lb_trans[2], precision=3, separator=","),
        numpy.array2string(center_trans[2], precision=3, separator=","),
        numpy.array2string(ub_trans[2], precision=3, separator=","),
    )

    return lb_trans[2], center_trans[2], ub_trans[2]


def flatten(chain):
    chain_shape = chain.shape
    flat_chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])
    return flat_chain


def change_bounds_json(cache, lb, ub, mcmc_store):
    "change the bounds based on lb and ub and then save it as a new json file and return the path to the new file"
    multiprocessing.get_logger().info("change_bounds_json  lb %s  ub %s", lb, ub)
    lb_trans = util.convert_individual_inputorder(lb, cache)
    ub_trans = util.convert_individual_inputorder(ub, cache)

    settings_file = Path(cache.json_path)
    settings_file_backup = settings_file.with_suffix(".json.backup")

    new_name = "%s_bounds%s" % (settings_file.stem, settings_file.suffix)

    new_settings_file = settings_file.with_name(new_name)

    with settings_file.open() as json_data:
        settings = jstyleson.load(json_data)

        idx = 0
        for parameter in settings["parameters"]:
            transform = cache.transforms[parameter["transform"]](parameter, cache)
            count = transform.count
            if count:
                multiprocessing.get_logger().warn("%s %s %s", idx, count, transform)
                lb_local = lb_trans[idx : idx + count]
                ub_local = ub_trans[idx : idx + count]
                transform.setBounds(parameter, lb_local, ub_local)
                idx = idx + count

        with new_settings_file.open(mode="w") as json_data:
            jstyleson.dump(settings, json_data, indent=4, sort_keys=False)

        mcmc_store.root.bounds_change.json = jstyleson.dumps(settings["parameters"], sort_keys=False)

    # copy the original file to a backup name
    shutil.copy(settings_file, settings_file_backup)

    # copy over our new settings file to the original file also
    # this is so that external programs also see the new bounds
    with settings_file.open(mode="w") as json_data:
        jstyleson.dump(settings, json_data, indent=4, sort_keys=False)

    return new_settings_file.as_posix()


def process_sampler_auto_bounds_write(cache, mcmc_store):
    bounds_seq = mcmc_store.root.bounds_acceptance
    bounds_chain = mcmc_store.root.bounds_full_chain

    bounds_chain, bounds_chain_flat, bounds_chain_transform, bounds_chain_flat_transform = process_chain(
        bounds_chain, cache, len(bounds_seq) - 1
    )

    mcmc_store.root.bounds_full_chain_transform = bounds_chain_transform
    mcmc_store.root.bounds_flat_chain = bounds_chain_flat
    mcmc_store.root.bounds_flat_chain_transform = bounds_chain_flat_transform

def select_best_kmeans(chain, probability):
    chain_shape = chain.shape
    flat_chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])
    flat_probability = numpy.squeeze(probability.reshape(-1, 1))

    #unique    
    flat_chain_unique, unique_indexes = numpy.unique(flat_chain, return_index=True, axis=0)
    flat_probability_unique = flat_probability[unique_indexes]
   
    #remove low probability
    flat_prob = numpy.exp(flat_probability_unique)
    max_prob = numpy.max(flat_prob)
    min_prob = max_prob/10  #10% of max prob cutoff
    
    selected = (flat_prob >= min_prob) & (flat_prob <= max_prob)
    
    flat_chain = flat_chain_unique[selected]
    flat_probability = flat_probability_unique[selected]

    if len(flat_chain) > (2* chain_shape[0]): 
        #kmeans clustering
        km = KMeans(chain_shape[0])
        km.fit(flat_chain)
    
        dist = scipy.spatial.distance.cdist(flat_chain, km.cluster_centers_)
    
        idx_closest = numpy.argmin(dist, 0)
    
        closest = dist[idx_closest, range(chain_shape[0])]

        best_chain = flat_chain[idx_closest]
        best_prob = flat_probability[idx_closest]
    else:
        pop_size = chain.shape[0]
        sort_idx = numpy.argsort(flat_probability_unique)
        sort_idx = sort_idx[numpy.isfinite(sort_idx)]

        best = sort_idx[-pop_size:]

        best_chain = flat_chain_unique[best,:]
        best_prob = flat_probability_unique[best]

    return best_chain, best_prob


def auto_high_probability(cache, checkpoint, sampler, iterations=100, steps=5):
    auto_chain = None
    auto_probability = None

    pop_size = checkpoint["populationSize"]

    for i in range(steps):
        chain, probability = auto_high_probability_iterations(cache, checkpoint, sampler, iterations)

        #store chain
        auto_chain = addChain(auto_chain, chain)
        auto_probability = addChain(auto_probability, probability)

        best_chain, best_prob = select_best_kmeans(auto_chain, auto_probability)

        multiprocessing.get_logger().info("best_chain %s", best_chain)
        multiprocessing.get_logger().info("best_prob %s", best_prob)

        checkpoint["p_bounds"] = best_chain
        checkpoint["ln_prob_bounds"] = best_prob
        checkpoint["rstate_bounds"] = None

    return auto_chain, auto_probability


def auto_high_probability_iterations(cache, checkpoint, sampler, iterations):
    auto_chain = None
    auto_probability = None
    best = -numpy.inf

    for i in range(iterations):
        state = next(
            sampler.sample(
                checkpoint["p_bounds"], log_prob0=checkpoint["ln_prob_bounds"], rstate0=checkpoint["rstate_bounds"], iterations=1
            )
        )

        p = state.coords
        ln_prob = state.log_prob
        random_state = state.random_state

        if any(ln_prob > best):
            best = numpy.max(ln_prob)

        accept = numpy.mean(sampler.acceptance_fraction)

        auto_chain = addChain(auto_chain, p[:, numpy.newaxis, :])
        auto_probability = addChain(auto_probability, ln_prob[:, numpy.newaxis])

        multiprocessing.get_logger().info("auto run: idx: %s accept: %.3f max ln(prob): %.3f", i, accept, best)

        checkpoint["p_bounds"] = p
        checkpoint["ln_prob_bounds"] = ln_prob
        checkpoint["rstate_bounds"] = random_state
    
    sampler.reset()
    return auto_chain, auto_probability


def sampler_auto_bounds(cache, checkpoint, sampler, checkpointFile, mcmc_store):
    bounds_seq = checkpoint.get("bounds_seq", [])

    bounds_chain = checkpoint.get("bounds_chain", None)
    bounds_probability = checkpoint.get("bounds_probability", None)

    checkInterval = 25

    parameters = len(cache.MIN_VALUE)

    if "mcmc_h5" in cache.settings:
        data = cadet.H5()
        data.filename = cache.settings["mcmc_h5"]
        data.load(paths=["/bounds_change/center_trans"])
        previous_parameters = data.root.bounds_change.center_trans.shape[1]
    else:
        previous_parameters = 0

    new_parameters = parameters - previous_parameters

    finished = False

    generation = checkpoint["idx_bounds"]

    sampler.iterations = checkpoint["sampler_iterations"]
    sampler.naccepted = checkpoint["sampler_naccepted"]
    sampler._moves[1].n = checkpoint["sampler_n"]

    auto_chain, auto_probability = auto_high_probability(cache, checkpoint, sampler, iterations=100, steps=5)

    mcmc_store.root.bounds.auto_chain = auto_chain
    mcmc_store.root.bounds.auto_probability = auto_probability

    mcmc_store.save()

    while not finished:
        state = next(
            sampler.sample(
                checkpoint["p_bounds"], log_prob0=checkpoint["ln_prob_bounds"], rstate0=checkpoint["rstate_bounds"], iterations=1
            )
        )

        p = state.coords
        ln_prob = state.log_prob
        random_state = state.random_state

        accept = numpy.mean(sampler.acceptance_fraction)
        bounds_seq.append(accept)

        bounds_chain = addChain(bounds_chain, p[:, numpy.newaxis, :])
        bounds_probability = addChain(bounds_probability, ln_prob[:, numpy.newaxis])

        multiprocessing.get_logger().info("run:  idx: %s accept: %.3f", generation, accept)

        generation += 1

        checkpoint["p_bounds"] = p
        checkpoint["ln_prob_bounds"] = ln_prob
        checkpoint["rstate_bounds"] = random_state
        checkpoint["idx_bounds"] = generation
        checkpoint["bounds_chain"] = bounds_chain
        checkpoint['bounds_probability'] = bounds_probability
        checkpoint["bounds_seq"] = bounds_seq
        checkpoint["bounds_iterations"] = sampler.iterations
        checkpoint["bounds_naccepted"] = sampler.naccepted

        mcmc_store.root.bounds_acceptance = numpy.array(bounds_seq).reshape(-1, 1)
        mcmc_store.root.bounds_full_chain = bounds_chain
        mcmc_store.root.bounds_probability = bounds_probability
        mcmc_store.root.bounds_probability_flat = bounds_probability.reshape(-1, 1)

        write_interval(cache.checkpointInterval, cache, checkpoint, checkpointFile, mcmc_store, process_sampler_auto_bounds_write)
        util.graph_corner_process(cache, last=False)

        if generation % checkInterval == 0:
            converged, lb, mid, ub = converged_bounds(bounds_chain[:, :, :new_parameters], 200, 1e-3)

            if converged:
                finished = True

                write_interval(-1, cache, checkpoint, checkpointFile, mcmc_store, process_sampler_auto_bounds_write)

                new_min_value, center, new_max_value = rescale(
                    cache, lb, mid, ub, numpy.array(cache.MIN_VALUE), numpy.array(cache.MAX_VALUE), mcmc_store
                )

                p_burn, ln_prob = select_best_kmeans(bounds_chain, bounds_probability)

                p_burn_trans = util.convert_population_inputorder(p_burn, cache)

                multiprocessing.get_logger().info("before bounds conversion %s", p_burn_trans)
                multiprocessing.get_logger().info("before bounds conversion p_burn %s", p_burn)

                json_path = change_bounds_json(cache, new_min_value, new_max_value, mcmc_store)
                cache.resetTransform(json_path)
                sampler.log_prob_fn.args[0] = json_path
                sampler.log_prob_fn.args[1] = cache

                p_burn = numpy.array([util.convert_individual_inverse(i, cache) for i in p_burn_trans])

                p_burn_trans = util.convert_population_inputorder(p_burn, cache)

                multiprocessing.get_logger().info("after bounds conversion %s", p_burn_trans)
                multiprocessing.get_logger().info("after bounds conversion p_burn %s", p_burn)

                checkpoint["state"] = "burn_in"
                checkpoint["p_burn"] = p_burn
                checkpoint['ln_prob_burn'] = ln_prob
                checkpoint['rstate_burn'] = None
                checkpoint["starting_population"] = p_burn

                write_checkpoint(-1, checkpoint, checkpointFile)

                sampler.reset()
            else:
                multiprocessing.get_logger().info("bounds have not yet converged in gen %s", generation)
                

def process_interval(cache, mcmc_store, interval_chain, interval_chain_transform):
    mean = numpy.mean(interval_chain_transform, 0)
    labels = [5, 10, 50, 90, 95]
    percentile = numpy.percentile(interval_chain_transform, labels, 0)

    mcmc_store.root.percentile["mean"] = mean
    for idx, label in enumerate(labels):
        mcmc_store.root.percentile["percentile_%s" % label] = percentile[idx, :]

    flat_interval = interval(interval_chain, cache)
    flat_interval_transform = interval(interval_chain_transform, cache)

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    flat_interval.to_csv(mcmcDir / "percentile.csv")
    flat_interval_transform.to_csv(mcmcDir / "percentile_transform.csv")


def process_sampler_burn_write(cache, mcmc_store):
    train_chain = mcmc_store.root.train_full_chain
    burn_seq = mcmc_store.root.burn_seq
    train_chain_stat = mcmc_store.root.train_chain_stat

    train_chain, train_chain_flat, train_chain_transform, train_chain_flat_transform = process_chain(train_chain, cache, len(burn_seq) - 1)

    mcmc_store.root.train_full_chain_transform = train_chain_transform
    mcmc_store.root.train_flat_chain = train_chain_flat
    mcmc_store.root.train_flat_chain_transform = train_chain_flat_transform

    train_chain_stat, _, train_chain_stat_transform, _ = process_chain(train_chain_stat, cache, len(burn_seq) - 1)

    mcmc_store.root.train_chain_stat_transform = train_chain_stat_transform

    interval_chain = train_chain_flat
    interval_chain_transform = train_chain_flat_transform
    process_interval(cache, mcmc_store, interval_chain, interval_chain_transform)


def sampler_burn(cache, checkpoint, sampler, checkpointFile, mcmc_store):
    burn_seq = checkpoint.get("burn_seq", [])

    train_chain = checkpoint.get("train_chain", None)
    train_probability = checkpoint.get('train_probability', None)

    train_chain_stat = checkpoint.get("train_chain_stat", None)

    converge = checkpoint.get("converge")

    parameters = len(cache.MIN_VALUE)

    tol = 5e-4
    power = checkpoint["sampler_n"]
    distance = 1.0
    # distance_a = sampler.a
    stop_next = False
    finished = False

    generation = checkpoint["idx_burn"]

    sampler.iterations = checkpoint["sampler_iterations"]
    sampler.naccepted = checkpoint["sampler_naccepted"]
    sampler._moves[1].n = checkpoint["sampler_n"]

    auto_chain, auto_probability = auto_high_probability(cache, checkpoint, sampler, iterations=100, steps=5)

    mcmc_store.root.burn.auto_chain = auto_chain
    mcmc_store.root.burn.auto_probability = auto_probability

    mcmc_store.save()

    while not finished:
        state = next(
            sampler.sample(
                checkpoint["p_burn"], log_prob0=checkpoint["ln_prob_burn"], rstate0=checkpoint["rstate_burn"], iterations=1, tune=False
            )
        )

        p = state.coords
        ln_prob = state.log_prob
        random_state = state.random_state

        accept = numpy.mean(sampler.acceptance_fraction)
        burn_seq.append(accept)
        converge[:-1] = converge[1:]
        converge[-1] = accept

        train_chain = addChain(train_chain, p[:, numpy.newaxis, :])
        train_probability = addChain(train_probability, ln_prob[:, numpy.newaxis])

        train_chain_stat = addChain(train_chain_stat, numpy.percentile(flatten(train_chain), [5, 50, 95], 0)[:, numpy.newaxis, :])

        converge_real = converge[~numpy.isnan(converge)]
        multiprocessing.get_logger().info(
            "burn:  idx: %s accept: %.3g std: %.3g mean: %.3g converge: %.3g",
            generation,
            accept,
            numpy.std(converge_real),
            numpy.mean(converge_real),
            numpy.std(converge_real) / tol,
        )

        generation += 1

        checkpoint["p_burn"] = p
        checkpoint["ln_prob_burn"] = ln_prob
        checkpoint["rstate_burn"] = random_state
        checkpoint["idx_burn"] = generation
        checkpoint["train_chain"] = train_chain
        checkpoint['train_probability'] = train_probability
        checkpoint["burn_seq"] = burn_seq
        checkpoint["converge"] = converge
        checkpoint["sampler_iterations"] = sampler.iterations
        checkpoint["sampler_naccepted"] = sampler.naccepted
        checkpoint["train_chain_stat"] = train_chain_stat
        checkpoint["sampler_n"] = sampler._moves[1].n

        mcmc_store.root.train_full_chain = train_chain
        mcmc_store.root.burn_seq = numpy.array(burn_seq).reshape(-1, 1)
        mcmc_store.root.train_chain_stat = train_chain_stat
        mcmc_store.root.train_probability = train_probability
        mcmc_store.root.train_probability_flat = train_probability.reshape(-1, 1)

        write_interval(cache.checkpointInterval, cache, checkpoint, checkpointFile, mcmc_store, process_sampler_burn_write)
        util.graph_corner_process(cache, last=False)

        if numpy.std(converge_real) < tol and len(converge) == len(converge_real):
            average_converge = numpy.mean(converge)
            if average_converge > min_acceptance:
                multiprocessing.get_logger().info("burn in completed at iteration %s", generation)
                finished = True

            if stop_next is True:
                multiprocessing.get_logger().info("burn in completed at iteration %s based on minimum distances", generation)
                finished = True

            if not finished:
                new_distance = min_acceptance - average_converge
                if new_distance < distance:
                    distance_n = sampler._moves[1].n
                    distance = new_distance

                    multiprocessing.get_logger().info(
                        "burn in acceptance is out of tolerance and n must be adjusted while burn in continues"
                    )
                    converge[:] = numpy.nan
                    prev_n = sampler._moves[1].n
                    if average_converge < (min_acceptance - 2 * acceptance_delta):
                        # n must be decreased to increase the acceptance rate (step size)
                        power -= 3
                    elif average_converge < (min_acceptance - acceptance_delta):
                        # n must be decreased to increase the acceptance rate (step size)
                        power -= 2
                    else:
                        power -= 1
                    new_n = power
                    sampler._moves[1].n = power

                    mcmc_store.root.train_power = power

                    sampler.reset()
                    checkpoint["p_burn"] = checkpoint["starting_population"]
                    checkpoint["ln_prob_burn"] = None
                    multiprocessing.get_logger().info("previous n: %s    new n: %s", prev_n, new_n)
                else:
                    sampler._moves[1].n = distance_n

                    mcmc_store.root.train_power = distance_n

                    sampler.reset()
                    checkpoint["p_burn"] = checkpoint["starting_population"]
                    checkpoint["ln_prob_burn"] = None
                    stop_next = True

    checkpoint["sampler_iterations"] = sampler.iterations
    checkpoint["sampler_naccepted"] = sampler.naccepted
    checkpoint["state"] = "chain"
    checkpoint["p_chain"] = p
    checkpoint["ln_prob_burn"] = ln_prob
    checkpoint["ln_prob_chain"] = ln_prob
    checkpoint["rstate_burn"] = random_state
    checkpoint["rstate_chain"] = random_state
    checkpoint["sampler_a"] = sampler._moves[1].n

    write_interval(-1, cache, checkpoint, checkpointFile, mcmc_store, process_sampler_burn_write)


def process_sampler_run_write(cache, mcmc_store):
    chain = mcmc_store.root.full_chain
    chain_seq = mcmc_store.root.mcmc_acceptance
    run_chain_stat = mcmc_store.root.run_chain_stat

    chain, chain_flat, chain_transform, chain_flat_transform = process_chain(chain, cache, len(chain_seq) - 1)

    mcmc_store.root.full_chain_transform = chain_transform
    mcmc_store.root.flat_chain = chain_flat
    mcmc_store.root.flat_chain_transform = chain_flat_transform

    run_chain_stat, _, run_chain_stat_transform, _ = process_chain(run_chain_stat, cache, len(chain_seq) - 1)

    mcmc_store.root.run_chain_stat_transform = run_chain_stat_transform

    interval_chain = chain_flat
    interval_chain_transform = chain_flat_transform
    process_interval(cache, mcmc_store, interval_chain, interval_chain_transform)


def sampler_run(cache, checkpoint, sampler, checkpointFile, mcmc_store):
    chain_seq = checkpoint.get("chain_seq", [])

    run_chain = checkpoint.get("run_chain", None)
    run_probability = checkpoint.get('run_probability', None)

    run_chain_stat = checkpoint.get("run_chain_stat", None)

    iat = checkpoint.get("integrated_autocorrelation_time", [])

    checkInterval = 25

    parameters = len(cache.MIN_VALUE)

    finished = False

    generation = checkpoint["idx_chain"]

    sampler.iterations = checkpoint["sampler_iterations"]
    sampler.naccepted = checkpoint["sampler_naccepted"]
    sampler._moves[1].n = checkpoint["sampler_n"]
    tau_percent = None

    while not finished:
        state = next(
            sampler.sample(checkpoint["p_chain"], log_prob0=checkpoint["ln_prob_chain"], rstate0=checkpoint["rstate_chain"], iterations=1)
        )

        p = state.coords
        ln_prob = state.log_prob
        random_state = state.random_state

        accept = numpy.mean(sampler.acceptance_fraction)
        chain_seq.append(accept)

        run_chain = addChain(run_chain, p[:, numpy.newaxis, :])
        run_probability = addChain(run_probability, ln_prob[:, numpy.newaxis])

        run_chain_stat = addChain(run_chain_stat, numpy.percentile(flatten(run_chain), [5, 50, 95], 0)[:, numpy.newaxis, :])

        multiprocessing.get_logger().info("run:  idx: %s accept: %.3f", generation, accept)

        generation += 1

        checkpoint["p_chain"] = p
        checkpoint["ln_prob_chain"] = ln_prob
        checkpoint["rstate_chain"] = random_state
        checkpoint["idx_chain"] = generation
        checkpoint["run_chain"] = run_chain
        checkpoint['run_probability'] = run_probability
        checkpoint["chain_seq"] = chain_seq
        checkpoint["sampler_iterations"] = sampler.iterations
        checkpoint["sampler_naccepted"] = sampler.naccepted
        checkpoint["run_chain_stat"] = run_chain_stat

        mcmc_store.root.full_chain = run_chain
        mcmc_store.root.mcmc_acceptance = numpy.array(chain_seq).reshape(-1, 1)
        mcmc_store.root.run_chain_stat = run_chain_stat
        mcmc_store.root.run_probability = run_probability
        mcmc_store.root.run_probability_flat = run_probability.reshape(-1, 1)

        if generation % checkInterval == 0:
            try:
                tau = autocorr.integrated_time(numpy.swapaxes(run_chain, 0, 1), tol=cache.MCMCTauMult)
                multiprocessing.get_logger().info(
                    "Mean acceptance fraction: %s %0.3f tau: %s with shape: %s", generation, accept, tau, run_chain.shape
                )
                if numpy.any(numpy.isnan(tau)):
                    multiprocessing.get_logger().info("tau is NaN and clearly not complete %s", generation)
                else:
                    multiprocessing.get_logger().info("we have run long enough and can quit %s", generation)
                    finished = True
            except autocorr.AutocorrError as err:
                multiprocessing.get_logger().info(str(err))
                tau = err.tau
            multiprocessing.get_logger().info("Mean acceptance fraction: %s %0.3f tau: %s", generation, accept, tau)

            temp_iat = [generation]
            temp_iat.extend(tau)
            iat.append(temp_iat)
            checkpoint["integrated_autocorrelation_time"] = iat

            mcmc_store.root.integrated_autocorrelation_time = numpy.array(iat)

            tau = numpy.array(tau)
            tau_percent = generation / (tau * cache.MCMCTauMult)

            mcmc_store.root.tau_percent = tau_percent.reshape(-1, 1)

        write_interval(cache.checkpointInterval, cache, checkpoint, checkpointFile, mcmc_store, process_sampler_run_write)
        mle_process(last=False)
        util.graph_corner_process(cache, last=False)

    checkpoint["p_chain"] = p
    checkpoint["ln_prob_chain"] = ln_prob
    checkpoint["rstate_chain"] = random_state
    checkpoint["idx_chain"] = generation
    checkpoint["run_chain"] = run_chain
    checkpoint['run_probability'] = run_probability
    checkpoint["chain_seq"] = chain_seq
    checkpoint["state"] = "complete"

    write_interval(-1, cache, checkpoint, checkpointFile, mcmc_store, process_sampler_run_write)


def write_checkpoint(interval, checkpoint, checkpointFile):
    "write the checkpoint and mcmc data at most every n seconds"
    if "last_time" not in write_checkpoint.__dict__:
        write_checkpoint.last_time = time.time()

    if time.time() - write_checkpoint.last_time > interval:
        with checkpointFile.open("wb") as cp_file:
            pickle.dump(checkpoint, cp_file)

        write_checkpoint.last_time = time.time()

def write_interval(interval, cache, checkpoint, checkpointFile, mcmc_store, process_mcmc_store):
    "write the checkpoint and mcmc data at most every n seconds"
    if "last_time" not in write_interval.__dict__:
        write_interval.last_time = time.time()

    write_checkpoint(interval, checkpoint, checkpointFile)

    if time.time() - write_interval.last_time > interval:
        writeMCMC(cache, mcmc_store, process_mcmc_store)

        write_interval.last_time = time.time()


def run(cache, tools, creator):
    "run the parameter estimation"
    random.seed()
    checkpointFile = Path(cache.settings["resultsDirMisc"], cache.settings.get("checkpointFile", "check"))
    checkpoint = getCheckPoint(checkpointFile, cache)

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    mcmc_h5 = mcmcDir / "mcmc.h5"
    mcmc_store = cadet.H5()
    mcmc_store.filename = mcmc_h5.as_posix()

    if mcmc_h5.exists():
        mcmc_store.load()

    parameters = len(cache.MIN_VALUE)

    MCMCpopulationSet = cache.settings.get("MCMCpopulationSet", None)
    if MCMCpopulationSet is not None:
        populationSize = MCMCpopulationSet
    else:
        populationSize = parameters * cache.settings["MCMCpopulation"]

    # Population must be even
    populationSize = populationSize + populationSize % 2

    # due to emcee 3.0 and RedBlueMove there is a minimum population size to work correctly based on the number of paramters
    populationSize = max(parameters * 2, populationSize)

    if checkpoint["state"] == "start":
        multiprocessing.get_logger().info("setting up kde")
        kde, kde_scaler = kde_generator.setupKDE(cache)
        checkpoint["state"] = "auto_bounds"

        with checkpointFile.open("wb") as cp_file:
            pickle.dump(checkpoint, cp_file)
    else:
        multiprocessing.get_logger().info("loading kde")
        kde, kde_scaler = kde_generator.getKDE(cache)

    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    with path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)

        result_data = {
            "input": [],
            "output": [],
            "output_meta": [],
            "results": {},
            "times": {},
            "input_transform": [],
            "input_transform_extended": [],
            "strategy": [],
            "mean": [],
            "confidence": [],
            "mcmc_score": [],
        }
        halloffame = pareto.DummyFront()
        meta_hof = pareto.ParetoFrontMeta(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache), slice_object=cache.meta_slice)
        grad_hof = pareto.DummyFront()
        progress_hof = None

        sampler = emcee.EnsembleSampler(
            populationSize,
            parameters,
            log_posterior_vectorize,
            args=[cache.json_path, cache, halloffame, meta_hof, grad_hof, progress_hof, result_data, writer, csvfile],
            moves=[(de_snooker.DESnookerMove(), 0.1), (de.DEMove(), 0.9 * 0.9), (de.DEMove(gamma0=1.0), 0.9 * 0.1),],
            vectorize=True,
        )

        if "sampler_n" not in checkpoint:
            checkpoint["sampler_n"] = sampler._moves[1].n

        if checkpoint["state"] == "auto_bounds":
            sampler_auto_bounds(cache, checkpoint, sampler, checkpointFile, mcmc_store)
            util.graph_corner_process(cache, last=False, interval=60)

        if checkpoint["state"] == "burn_in":
            sampler_burn(cache, checkpoint, sampler, checkpointFile, mcmc_store)
            util.graph_corner_process(cache, last=False, interval=60)

        if checkpoint["state"] == "chain":
            run_chain = checkpoint.get("run_chain", None)
            if run_chain is not None:
                temp = run_chain[:, : checkpoint["idx_chain"], 0].T
                multiprocessing.get_logger().info("complete shape %s", temp.shape)

                try:
                    tau = autocorr.integrated_time(numpy.swapaxes(run_chain[:, : checkpoint["idx_chain"], :], 0, 1), tol=cache.MCMCTauMult)
                    multiprocessing.get_logger().info("we have previously run long enough and can quit %s", checkpoint["idx_chain"])
                    checkpoint["state"] = "complete"
                except autocorr.AutocorrError as err:
                    multiprocessing.get_logger().info(str(err))
                    tau = err.tau

                multiprocessing.get_logger().info(
                    "Mean acceptance fraction: %s %0.3f tau: %s", checkpoint["idx_chain"], checkpoint["chain_seq"][-1], tau
                )

        if checkpoint["state"] == "chain":
            sampler_run(cache, checkpoint, sampler, checkpointFile, mcmc_store)
            util.graph_corner_process(cache, last=False, interval=60)

    chain = checkpoint["run_chain"]
    chain_shape = chain.shape
    chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])

    if checkpoint["state"] == "complete":
        util.finish(cache)
        checkpoint["state"] = "plot_finish"

        with checkpointFile.open("wb") as cp_file:
            pickle.dump(checkpoint, cp_file)

    if checkpoint["state"] == "plot_finish":
        mle_process(last=True)
        tube_process(last=True)
        util.graph_corner_process(cache, last=True)
    return numpy.mean(chain, 0)


def tube_process(last=False, interval=3600):
    cwd = str(Path(__file__).parent.parent)
    ret = subprocess.run(
        [sys.executable, "mcmc_plot_tube.py", str(cache.cache.json_path), str(util.getCoreCounts())],
        stdin=None,
        stdout=None,
        stderr=None,
        close_fds=True,
        cwd=cwd,
    )


def mle_process(last=False, interval=3600*24):
    if "last_time" not in mle_process.__dict__:
        mle_process.last_time = time.time()

    if "child" in mle_process.__dict__:
        if mle_process.child.poll() is None:  # This is false if the child has completed
            if last:
                mle_process.child.wait()
            else:
                return

    cwd = str(Path(__file__).parent.parent)

    if last:
        ret = subprocess.run(
            [sys.executable, "mle.py", str(cache.cache.json_path), str(util.getCoreCounts())],
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
            cwd=cwd,
        )
        mle_process.last_time = time.time()
    elif (time.time() - mle_process.last_time) > interval:
        subprocess.run(
            [sys.executable, "mle.py", str(cache.cache.json_path), str(util.getCoreCounts())],
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
            cwd=cwd,
        )
        mle_process.last_time = time.time()


def get_population(base, size, diff=0.02):
    new_population = base
    row, col = base.shape
    multiprocessing.get_logger().info("%s", base)

    multiprocessing.get_logger().info("row %s size %s", row, size)
    if row < size:
        # create new entries
        indexes = numpy.random.choice(new_population.shape[0], size - row, replace=True)
        temp = new_population[indexes, :]
        rand = numpy.random.normal(1.0, diff, size=temp.shape)
        new_population = numpy.concatenate([new_population, temp * rand])
    if row > size:
        # randomly select entries to keep
        indexes = numpy.random.choice(new_population.shape[0], size, replace=False)
        multiprocessing.get_logger().info("indexes: %s", indexes)
        new_population = new_population[indexes, :]
    return new_population


def resetPopulation(checkpoint, cache):
    populationSize = checkpoint["populationSize"]
    parameters = len(cache.MIN_VALUE)

    if cache.settings.get("PreviousResults", None) is not None:
        multiprocessing.get_logger().info("running with previous best results")
        previousResultsFile = Path(cache.settings["PreviousResults"])
        results_h5 = cadet.H5()
        results_h5.filename = previousResultsFile.as_posix()
        results_h5.load()
        previousResults = results_h5.root.meta_population_transform

        row, col = previousResults.shape
        multiprocessing.get_logger().info("row: %s col: %s  parameters: %s", row, col, parameters)
        if col < parameters:
            mcmc_h5 = Path(cache.settings.get("mcmc_h5", None))
            mcmcDir = mcmc_h5.parent
            mle_h5 = mcmcDir / "mle.h5"

            data = cadet.H5()
            data.filename = mle_h5.as_posix()
            data.load()
            multiprocessing.get_logger().info("%s", list(data.root.keys()))
            stat_MLE = data.root.stat_MLE.reshape(1, -1)
            previousResults = numpy.hstack([previousResults, numpy.repeat(stat_MLE, row, 0)])
            multiprocessing.get_logger().info("row: %s  col:%s   shape: %s", row, col, previousResults.shape)

        population = get_population(previousResults, populationSize, diff=0.02)
        checkpoint["starting_population"] = population
        checkpoint["starting_population"] = [util.convert_individual_inverse(i, cache) for i in population]
        multiprocessing.get_logger().info("p_burn startup population: %s", population)
        multiprocessing.get_logger().info("p_burn startup: %s", checkpoint["starting_population"])
    else:
        checkpoint["starting_population"] = SALib.sample.sobol_sequence.sample(populationSize, parameters)
    checkpoint["p_bounds"] = checkpoint["starting_population"]


def getCheckPoint(checkpointFile, cache):
    if checkpointFile.exists():
        with checkpointFile.open("rb") as cp_file:
            checkpoint = pickle.load(cp_file)
    else:
        parameters = len(cache.MIN_VALUE)

        MCMCpopulationSet = cache.settings.get("MCMCpopulationSet", None)
        if MCMCpopulationSet is not None:
            populationSize = MCMCpopulationSet
        else:
            populationSize = parameters * cache.settings["MCMCpopulation"]

        # Population must be even
        populationSize = populationSize + populationSize % 2

        # due to emcee 3.0 and RedBlueMove there is a minimum population size to work correctly based on the number of paramters
        populationSize = max(parameters * 2, populationSize)

        checkpoint = {}
        checkpoint["state"] = "start"
        checkpoint["populationSize"] = populationSize
        resetPopulation(checkpoint, cache)

        checkpoint["ln_prob_bounds"] = None
        checkpoint["rstate_bounds"] = None
        checkpoint["idx_bounds"] = 0

        checkpoint["ln_prob_burn"] = None
        checkpoint["rstate_burn"] = None
        checkpoint["idx_burn"] = 0

        checkpoint["p_chain"] = None
        checkpoint["ln_prob_chain"] = None
        checkpoint["rstate_chain"] = None
        checkpoint["idx_chain"] = 0

        checkpoint["sampler_iterations"] = 0
        checkpoint["sampler_naccepted"] = numpy.zeros(populationSize)

        checkpoint["converge"] = numpy.ones(cache.settings.get("burnStable", 50)) * numpy.nan

    checkpoint["length_chain"] = cache.settings.get("chainLength", 50000)
    checkpoint["length_burn"] = cache.settings.get("burnIn", 50000)
    return checkpoint


def setupDEAP(cache, fitness, fitness_final, grad_fitness, grad_search, grad_search_fine, map_function, creator, base, tools):
    "setup the DEAP variables"
    creator.create("FitnessMax", base.Fitness, weights=[1.0] * cache.numGoals)
    creator.create("Individual", array.array, typecode="d", fitness=creator.FitnessMax, strategy=None, mean=None, confidence=None)

    creator.create("FitnessMaxMeta", base.Fitness, weights=[1.0, 1.0, 1.0, -1.0, -1.0])
    creator.create("IndividualMeta", array.array, typecode="d", fitness=creator.FitnessMaxMeta, strategy=None, best=None)
    cache.toolbox.register("individualMeta", util.initIndividual, creator.IndividualMeta, cache)

    cache.toolbox.register(
        "individual", util.generateIndividual, creator.Individual, len(cache.MIN_VALUE), cache.MIN_VALUE, cache.MAX_VALUE, cache
    )

    cache.toolbox.register("individual_guess", util.initIndividual, creator.Individual, cache)

    cache.toolbox.register("map", map_function)


def process(population_order, cache, halloffame, meta_hof, grad_hof, progress_hof, result_data, results, writer, csv_file):
    if "gen" not in process.__dict__:
        process.gen = 0

    if "sim_start" not in process.__dict__:
        process.sim_start = time.time()

    if "generation_start" not in process.__dict__:
        process.generation_start = time.time()

    population_lookup = {}
    fitnesses_lookup = {}

    log_likelihoods_lookup = {}

    keep = set()

    for ll, theta, fit, csv_line, meta_score, result, individual in results:
        log_likelihoods_lookup[tuple(individual)] = float(ll)
        if len(csv_line):
            #if csv_line is blank it means a simulation failed or is out of bounds, if this is handed on for
            #processing it will cause problems, failed simulations are arleady recorded and we don't want their
            #failure data in the recorers where other stuff picks it up
            keep.add(tuple(individual))            

            fitnesses_lookup[tuple(individual)] = (fit, csv_line, meta_score, result, tuple(individual))

            ind = cache.toolbox.individual_guess(individual)
            population_lookup[tuple(individual)] = ind

    # everything above is async (unordered) and needs to be reordered based on the population_order
    population = [population_lookup[tuple(row)] for row in population_order if tuple(row) in keep]
    fitnesses = [fitnesses_lookup[tuple(row)] for row in population_order if tuple(row) in keep]
    log_likelihoods = [log_likelihoods_lookup[tuple(row)] for row in population_order if tuple(row) in keep]
    log_likelihoods_all = [log_likelihoods_lookup[tuple(row)] for row in population_order]

    stalled, stallWarn, progressWarn = util.process_population(
        cache.toolbox, cache, population, fitnesses, writer, csv_file, halloffame, meta_hof, progress_hof, process.gen, result_data
    )

    progress.writeProgress(
        cache,
        process.gen,
        population,
        halloffame,
        meta_hof,
        grad_hof,
        progress_hof,
        process.sim_start,
        process.generation_start,
        result_data,
        line_log=False,
        probability=numpy.exp(log_likelihoods)
    )

    util.graph_process(cache, process.gen)

    process.gen += 1
    process.generation_start = time.time()
    return log_likelihoods_all


def process_chain(chain, cache, idx):
    chain_shape = chain.shape
    flat_chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])

    flat_chain_transform = util.convert_population_inputorder(flat_chain, cache)
    chain_transform = flat_chain_transform.reshape(chain_shape)

    return chain, flat_chain, chain_transform, flat_chain_transform


def writeMCMC(cache, mcmc_store, process_mcmc_store):
    "write out the mcmc data so it can be plotted"
    process_mcmc_store(cache, mcmc_store)
    mcmc_store.save()


def interval(flat_chain, cache):
    mean = numpy.mean(flat_chain, 0)

    percentile = numpy.percentile(flat_chain, [5, 10, 50, 90, 95], 0)

    data = numpy.vstack((mean, percentile)).transpose()

    pd = pandas.DataFrame(data, columns=["mean", "5", "10", "50", "90", "95"])
    pd.insert(0, "name", cache.parameter_headers_actual)
    pd.set_index("name")
    return pd



import sys

import matplotlib
import matplotlib.style as mplstyle
mplstyle.use('fast')

matplotlib.use("Agg")

size = 20

matplotlib.rc("font", size=size)  # controls default text sizes
matplotlib.rc("axes", titlesize=size)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=size)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=size)  # legend fontsize
matplotlib.rc("figure", titlesize=size)  # fontsize of the figure title
matplotlib.rc("figure", autolayout=True)

from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from mpl_toolkits.mplot3d import Axes3D

from CADETMatch.cache import cache

from pathlib import Path
import numpy
import scipy.interpolate
import itertools

from sklearn.preprocessing import MinMaxScaler

from cadet import Cadet, H5
from addict import Dict

# parallelization
import multiprocessing

import os
import warnings
import CADETMatch.util as util
import logging
import CADETMatch.loggerwriter as loggerwriter

from matplotlib.colors import ListedColormap
import matplotlib.cm

cm_plot = matplotlib.cm.gist_rainbow

def get_color(idx, max_colors, cmap):
    return cmap(1.0 * float(idx) / max_colors)

def main(map_function):
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "mixing.log")
    cache.setup(sys.argv[1])

    multiprocessing.get_logger().info("mixing graphing directory %s", os.getcwd())

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    mcmc_h5 = mcmcDir / "mcmc.h5"
    if mcmc_h5.exists():
        mcmc_store = H5()
        mcmc_store.filename = mcmc_h5.as_posix()
        mcmc_store.load(paths=["/full_chain", "/train_full_chain", "/bounds_full_chain",
                               "/full_chain_transform", "/train_full_chain_transform", "/bounds_full_chain_transform",
                               "/bounds_probability", "/train_probability", "/run_probability"])

        progress_path = Path(cache.settings["resultsDirBase"]) / "result.h5"

        graph_dir = cache.settings["resultsDirSpace"] / "mcmc"

        mixing = graph_dir / "mixing"
        mixing.mkdir(parents=True, exist_ok=True)

        input_headers = cache.parameter_headers_actual

        chain_names = (("full_chain", "run_probability"), 
                       ("train_full_chain", "train_probability"), 
                       ("bounds_full_chain", "bounds_probability"), 
                       ("full_chain_transform", "run_probability"), 
                       ("train_full_chain_transform", "train_probability"),  
                       ("bounds_full_chain_transform", "bounds_probability"))

        for chain, probability in chain_names:
            if chain in mcmc_store.root:
                plot_chain(input_headers, mcmc_store.root[chain], mcmc_store.root[probability], mixing / ("mixing_%s" % chain))


def plot_chain(headers, chain, probability, graph_dir):
    graph_dir.mkdir(parents=True, exist_ok=True)

    for i in range(chain.shape[2]):
        graph_mix(headers[i], chain[:, :, i], graph_dir)
        graph_probability(headers[i], chain[:, :, i], probability, graph_dir)
        

def graph_mix(header, chain, graph_dir):
    fig = figure.Figure(figsize=[15, 7])
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)

    chain_length = chain.shape[1]
    x = numpy.linspace(0, chain_length -1, chain_length)

    lines = []
    for j in range(chain.shape[0]):
        graph.plot(x, chain[j, :], color = get_color(j, chain.shape[0] - 1, cm_plot))

    graph.set_xlabel("chain length")
    graph.set_ylabel("value")
    filename = "mixing_%s.png" % (header)
    filename = filename.replace(":", "_").replace("/", "_")

    graph.set_title("Mixing graph %s" % header)
    fig.savefig((graph_dir / filename).as_posix())


def graph_probability(header, chain, probability, graph_dir):
    scaler = MinMaxScaler()
    probability = scaler.fit_transform(probability)

    fig = figure.Figure(figsize=[15, 7])
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)

    chain_length = chain.shape[1]
    x = numpy.linspace(0, chain_length -1, chain_length)

    lines = []
    for j in range(chain.shape[0]):
        graph.scatter(x, chain[j, :], color = matplotlib.cm.rainbow(probability[j,:]))

    graph.set_xlabel("chain length")
    graph.set_ylabel("value")
    filename = "mixing_%s_probability.png" % (header)
    filename = filename.replace(":", "_").replace("/", "_")

    graph.set_title("Mixing graph %s" % header)
    fig.savefig((graph_dir / filename).as_posix())

if __name__ == "__main__":
    map_function = util.getMapFunction()
    main(map_function)
    sys.exit()



import CADETMatch.util as util
import CADETMatch.score as score
import scipy.stats
from addict import Dict
import numpy

name = "similarityDecay"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 0
    settings.meta_mask = True
    settings.count = 3
    return settings


def run(sim_data, feature):
    "similarity, value, start stop"
    sim_time_values, sim_data_values = util.get_times_values(sim_data["simulation"], feature)
    selected = feature["selected"]

    exp_data_values = feature["value"][selected]
    exp_time_values = feature["time"][selected]

    [high, low] = util.find_peak(exp_time_values, sim_data_values)

    time_high, value_high = high

    pearson, diff_time = score.pearson_spline(exp_time_values, sim_data_values, feature["smooth_value"])

    temp = [pearson, feature["value_function"](value_high), feature["time_function"](numpy.abs(diff_time))]
    return (
        temp,
        util.sse(sim_data_values, exp_data_values),
        len(sim_data_values),
        sim_time_values,
        sim_data_values,
        exp_data_values,
        [1.0 - i for i in temp],
    )


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    name = "%s_%s" % (sim.root.experiment_name, feature["name"])
    s, crit_fs, crit_fs_der = smoothing.find_smoothing_factors(selectedTimes, selectedValues, name, cache)

    temp = {}
    temp["peak"] = util.find_peak(selectedTimes, selectedValues)[0]
    temp["time_function"] = score.time_function_decay(feature["time"][-1])
    temp["value_function"] = score.value_function(temp["peak"][1], abstol)
    temp["peak_max"] = max(selectedValues)
    temp["smoothing_factor"] = s
    temp["critical_frequency"] = crit_fs
    temp["critical_frequency_der"] = crit_fs_der
    temp["smooth_value"] = smoothing.smooth_data(selectedTimes, selectedValues, crit_fs, s)
    return temp


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    temp = ["%s_Similarity" % name, "%s_Value" % name, "%s_Time" % name]
    return temp



import CADETMatch.util as util
import CADETMatch.score as score
import numpy
from addict import Dict

name = "width"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 0
    settings.meta_mask = True
    settings.count = 3
    return settings


def run(sim_data, feature):
    "similarity, value, start stop"
    sim_time_values, sim_data_values = util.get_times_values(sim_data["simulation"], feature)
    selected = feature["selected"]

    exp_data_values = feature["value"][selected]
    exp_time_values = feature["time"][selected]

    sim_width_25 = find_width(sim_time_values, sim_data_values, 0.25)
    sim_width_50 = find_width(sim_time_values, sim_data_values, 0.50)
    sim_width_75 = find_width(sim_time_values, sim_data_values, 0.75)

    temp = [
        feature["width_25"](sim_width_25),
        feature["width_50"](sim_width_50),
        feature["width_75"](sim_width_75),
    ]
    return (
        temp,
        util.sse(sim_data_values, exp_data_values),
        len(sim_data_values),
        sim_time_values,
        sim_data_values,
        exp_data_values,
        [1.0 - i for i in temp],
    )


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    temp["width_25"] = score.value_function(find_width(selectedTimes, selectedValues, 0.25), abstol)
    temp["width_50"] = score.value_function(find_width(selectedTimes, selectedValues, 0.50), abstol)
    temp["width_75"] = score.value_function(find_width(selectedTimes, selectedValues, 0.75), abstol)
    temp["peak_max"] = max(selectedValues)
    return temp


def find_width(times, values, percent):
    idx_max = numpy.argmax(values)

    max_value = values[idx_max]
    max_time = times[idx_max]

    idx_upper = numpy.argmax(values[idx_max:] <= (max_value * percent))

    idx_lower = numpy.argmin(values[:idx_max] <= (max_value * percent))

    diff_time = times[idx_max + idx_upper] - times[idx_lower]

    return diff_time


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    temp = ["%s_width_25" % name, "%s_width_50" % name, "%s_width_75" % name]
    return temp



import numpy


def sse(exp_data, sim_data):
    return numpy.sum((numpy.array(exp_data) - numpy.array(sim_data)) ** 2)


def mse(exp_data, sim_data):
    return sse(exp_data, sim_data) / len(exp_data)


def rmse(exp_data, sim_data):
    return numpy.sqrt(mse(exp_data, sim_data))


def sse_norm(exp_data, sim_data):
    return numpy.sum((numpy.array(exp_data) - numpy.array(sim_data)) ** 2) / (numpy.max(exp_data) ** 2)


def mse_norm(exp_data, sim_data):
    return sse_norm(exp_data, sim_data) / len(exp_data)


def norm_rmse(exp_data, sim_data):
    return numpy.sqrt(mse_norm(exp_data, sim_data))


def rmse_combine(exp_datas, sim_datas):
    len_entries = numpy.array([len(exp_data) for exp_data in exp_datas])
    sse_s = numpy.array([sse_norm(exp_data, sim_data) for exp_data, sim_data in zip(exp_datas, sim_datas)])
    mse_s = sse_s / len_entries
    mean_total = numpy.sum(mse_s)
    rmse = numpy.sqrt(mean_total)
    return rmse



import CADETMatch.util as util
import random
from pathlib import Path
import csv
import CADETMatch.pareto as pareto
import CADETMatch.progress as progress
import multiprocessing
import time
import array

name = "GraphSpace"

"This is not actually an optimization method. It is used to sample a space and generate graphs to see what the space looks like."


def run(cache, tools, creator):
    "run the parameter estimation"
    random.seed()

    parameters = len(cache.MIN_VALUE)

    LAMBDA = parameters * cache.settings["population"]
    sim_start = generation_start = time.time()
    result_data = {
        "input": [],
        "output": [],
        "output_meta": [],
        "results": {},
        "times": {},
        "input_transform": [],
        "input_transform_extended": [],
        "strategy": [],
        "mean": [],
        "confidence": [],
    }

    pop = cache.toolbox.population(n=LAMBDA)

    if "seeds" in cache.settings:
        seed_pop = [
            cache.toolbox.individual_guess([f(v) for f, v in zip(cache.settings["transform"], sublist)])
            for sublist in cache.settings["seeds"]
        ]
        pop.extend(seed_pop)

    gradCheck = cache.badScore

    if cache.metaResultsOnly:
        hof = pareto.DummyFront()
    else:
        hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
    meta_hof = pareto.ParetoFrontMeta(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache), slice_object=cache.meta_slice)
    grad_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
    progress_hof = pareto.DummyFront()

    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    with path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)

        multiprocessing.get_logger().info("Population %s", pop)

        # gradCheck, newChildren = cache.toolbox.grad_search(gradCheck, pop, cache, writer, csvfile, hof, meta_hof, -1, check_all=True)

        stalled, stallWarn, progressWarn = util.eval_population(
            cache.toolbox, cache, pop, writer, csvfile, hof, meta_hof, None, -1, result_data
        )

        progress.writeProgress(cache, -1, pop, hof, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data)

        util.finish(cache)
        util.graph_corner_process(cache, last=True)

        return hof


def setupDEAP(cache, fitness, fitness_final, grad_fitness, grad_search, grad_search_fine, map_function, creator, base, tools):
    "setup the DEAP variables"
    creator.create("FitnessMax", base.Fitness, weights=[1.0] * cache.numGoals)
    creator.create("Individual", array.array, typecode="d", fitness=creator.FitnessMax, strategy=None, mean=None, confidence=None)

    creator.create("FitnessMaxMeta", base.Fitness, weights=[1.0, 1.0, 1.0, -1.0, -1.0])
    creator.create("IndividualMeta", array.array, typecode="d", fitness=creator.FitnessMaxMeta, strategy=None, best=None)
    cache.toolbox.register("individualMeta", util.initIndividual, creator.IndividualMeta, cache)

    cache.toolbox.register(
        "individual", util.generateIndividual, creator.Individual, len(cache.MIN_VALUE), cache.MIN_VALUE, cache.MAX_VALUE, cache
    )

    if cache.sobolGeneration:
        cache.toolbox.register("population", util.sobolGenerator, creator.Individual, cache)
    else:
        cache.toolbox.register("population", tools.initRepeat, list, cache.toolbox.individual)
    cache.toolbox.register("randomPopulation", tools.initRepeat, list, cache.toolbox.individual)

    cache.toolbox.register("individual_guess", util.initIndividual, creator.Individual, cache)

    cache.toolbox.register("evaluate", fitness, json_path=cache.json_path)
    cache.toolbox.register("evaluate_final", fitness_final, json_path=cache.json_path)
    cache.toolbox.register("evaluate_grad", grad_fitness, json_path=cache.json_path)
    cache.toolbox.register("evaluate_grad_fine", grad_search_fine, json_path=cache.json_path)
    cache.toolbox.register("grad_search", grad_search)

    cache.toolbox.register("map", map_function)



from pathlib import Path
import pickle
import random
import numpy
import CADETMatch.util as util
import CADETMatch.progress as progress
from deap import algorithms
import time
import csv
import CADETMatch.pareto as pareto
import multiprocessing

stallRate = 1.25
progressRate = 0.75

def stop_iteration(best, stalled, cache):
    best = util.translate_meta_min(best, cache)
    stopAverage = cache.settings.get("stopAverage", 0.0)
    stopBest = cache.settings.get("stopBest", 0.0)
    stopRMSE = cache.settings.get("stopRMSE", 0.0)
    if best[2] <= stopAverage or best[1] <= stopBest or stalled or best[-1] <= stopRMSE:
        return True
    else:
        return False

def eaMuPlusLambda(population, toolbox, mu, lambda_, cxpb, mutpb, ngen, settings, stats=None, verbose=__debug__, tools=None, cache=None):
    """from DEAP function but with checkpoiting"""
    assert lambda_ >= mu, "lambda must be greater or equal to mu."

    checkpointFile = Path(settings["resultsDirMisc"], settings.get("checkpointFile", "check"))

    sim_start = generation_start = time.time()

    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    result_data = {
        "input": [],
        "output": [],
        "output_meta": [],
        "results": {},
        "times": {},
        "input_transform": [],
        "input_transform_extended": [],
        "strategy": [],
        "mean": [],
        "confidence": [],
    }
    with path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)

        if checkpointFile.exists():
            with checkpointFile.open("rb") as cp_file:
                cp = pickle.load(cp_file)
            population = cp["population"]
            start_gen = cp["generation"]

            halloffame = cp["halloffame"]
            meta_hof = cp["meta_halloffame"]
            grad_hof = cp["grad_halloffame"]
            progress_hof = cp["progress_halloffame"]
            random.setstate(cp["rndstate"])
            cache.generationsOfProgress = cp["generationsOfProgress"]
            cache.lastProgressGeneration = cp["lastProgressGeneration"]

            if cp["gradCheck"] > cache.settings.get("gradCheck", 0.0):
                gradCheck = cp["gradCheck"]
            else:
                gradCheck = cache.settings.get("gradCheck", 0.0)
        else:
            # Start a new evolution
            start_gen = 0

            gradCheck = settings.get("gradCheck", 0.0)

            if cache.metaResultsOnly:
                halloffame = pareto.DummyFront()
            else:
                halloffame = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
            meta_hof = pareto.ParetoFrontMeta(
                similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache), slice_object=cache.meta_slice
            )
            grad_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
            progress_hof = pareto.ParetoFrontMeta(
                similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache), slice_object=cache.meta_slice
            )

            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in population if not ind.fitness.valid]
            if invalid_ind:
                stalled, stallWarn, progressWarn = util.eval_population(
                    toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, progress_hof, -1, result_data
                )

                progress.writeProgress(
                    cache, -1, population, halloffame, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data
                )
                util.graph_process(cache, "First")
                util.graph_corner_process(cache, last=False)

            cp = dict(
                population=population,
                generation=start_gen,
                halloffame=halloffame,
                rndstate=random.getstate(),
                gradCheck=gradCheck,
                meta_halloffame=meta_hof,
                grad_halloffame=grad_hof,
                generationsOfProgress=cache.generationsOfProgress,
                lastProgressGeneration=cache.lastProgressGeneration,
                progress_halloffame=progress_hof,
            )

            with checkpointFile.open("wb") as cp_file:
                pickle.dump(cp, cp_file)

        gen = start_gen  # this covers the case where the start_gen is higher than our stop gen so the loop never runs
        # Begin the generational process
        for gen in range(start_gen, ngen + 1):
            generation_start = time.time()
            # Vary the population
            offspring = algorithms.varAnd(population, toolbox, cxpb, mutpb)

            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
            stalled, stallWarn, progressWarn = util.eval_population(
                toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, progress_hof, gen, result_data
            )

            gradCheck, newChildren = cache.toolbox.grad_search(gradCheck, offspring, cache, writer, csvfile, grad_hof, meta_hof, gen)
            offspring.extend(newChildren)

            # Select the next generation population
            population[:] = toolbox.select(offspring + population, mu)

            progress.writeProgress(
                cache, gen, offspring, halloffame, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data
            )
            util.graph_process(cache, gen)
            util.graph_corner_process(cache, last=False)

            cp = dict(
                population=population,
                generation=gen,
                halloffame=halloffame,
                rndstate=random.getstate(),
                gradCheck=gradCheck,
                meta_halloffame=meta_hof,
                grad_halloffame=grad_hof,
                generationsOfProgress=cache.generationsOfProgress,
                lastProgressGeneration=cache.lastProgressGeneration,
                progress_halloffame=progress_hof,
            )

            if stallWarn:
                maxPopulation = cache.settings["maxPopulation"] * len(cache.MIN_VALUE)
                newLambda_ = int(lambda_ * stallRate)
                lambda_ = min(newLambda_, maxPopulation)

            if progressWarn:
                minPopulation = cache.settings["minPopulation"] * len(cache.MIN_VALUE)
                newLambda_ = int(lambda_ * progressRate)
                lambda_ = max(newLambda_, minPopulation)

            with checkpointFile.open("wb") as cp_file:
                pickle.dump(cp, cp_file)

            best = meta_hof.getBestScores()

            if stop_iteration(best, stalled, cache):
                break

        if cache.finalGradRefinement:
            gen = gen + 1
            best_individuals = [cache.toolbox.individual_guess(i) for i in meta_hof]
            gradCheck, newChildren = cache.toolbox.grad_search(
                gradCheck, best_individuals, cache, writer, csvfile, grad_hof, meta_hof, gen, check_all=True, result_data=result_data
            )
            if newChildren:
                progress.writeProgress(
                    cache, gen, newChildren, halloffame, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data
                )

        population = [cache.toolbox.individual_guess(i) for i in meta_hof]
        stalled, stallWarn, progressWarn = util.eval_population_final(
            cache.toolbox, cache, population, writer, csvfile, halloffame, meta_hof, progress_hof, gen + 1, result_data
        )
        progress.writeProgress(
            cache, gen + 1, population, halloffame, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data
        )

        util.finish(cache)
        util.graph_corner_process(cache, last=True)
        return halloffame


def nsga2(populationSize, ngen, cache, tools):
    """NSGA2 with checkpointing"""
    # fix max population to be a multiple of 4
    cache.settings["maxPopulation"] = cache.settings["maxPopulation"] + (-cache.settings["maxPopulation"] % 4)

    cxpb = cache.settings["crossoverRate"]
    checkpointFile = Path(cache.settings["resultsDirMisc"], cache.settings.get("checkpointFile", "check"))

    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    result_data = {
        "input": [],
        "output": [],
        "output_meta": [],
        "results": {},
        "times": {},
        "input_transform": [],
        "input_transform_extended": [],
        "strategy": [],
        "mean": [],
        "confidence": [],
    }
    with path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)

        if checkpointFile.exists():
            with checkpointFile.open("rb") as cp_file:
                cp = pickle.load(cp_file)
            population = cp["population"]
            start_gen = cp["generation"]

            halloffame = cp["halloffame"]
            meta_hof = cp["meta_halloffame"]
            grad_hof = cp["grad_halloffame"]
            progress_hof = cp["progress_halloffame"]
            random.setstate(cp["rndstate"])
            cache.generationsOfProgress = cp["generationsOfProgress"]
            cache.lastProgressGeneration = cp["lastProgressGeneration"]

            if cp["gradCheck"] > cache.settings.get("gradCheck", 0.0):
                gradCheck = cp["gradCheck"]
            else:
                gradCheck = cache.settings.get("gradCheck", 0.0)

        else:
            # Start a new evolution

            population = cache.toolbox.population(n=populationSize)

            if "seeds" in cache.settings:
                seed_pop = [
                    cache.toolbox.individual_guess([f(v) for f, v in zip(cache.settings["transform"], sublist)])
                    for sublist in cache.settings["seeds"]
                ]
                population.extend(seed_pop)

            start_gen = 0

            if cache.metaResultsOnly:
                halloffame = pareto.DummyFront()
            else:
                halloffame = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
            meta_hof = pareto.ParetoFrontMeta(
                similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache), slice_object=cache.meta_slice
            )
            grad_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
            progress_hof = pareto.ParetoFrontMeta(
                similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache), slice_object=cache.meta_slice
            )
            gradCheck = cache.settings.get("gradCheck", 0.0)

        sim_start = generation_start = time.time()

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in population if not ind.fitness.valid]
        if invalid_ind:
            stalled, stallWarn, progressWarn = util.eval_population(
                cache.toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, progress_hof, -1, result_data
            )

            progress.writeProgress(
                cache, -1, population, halloffame, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data
            )
            util.graph_process(cache, "First")
            util.graph_corner_process(cache, last=False)

        # This is just to assign the crowding distance to the individuals
        # no actual selection is done
        population = cache.toolbox.select(population, len(population))

        cp = dict(
            population=population,
            generation=start_gen,
            halloffame=halloffame,
            rndstate=random.getstate(),
            gradCheck=gradCheck,
            meta_halloffame=meta_hof,
            grad_halloffame=grad_hof,
            generationsOfProgress=cache.generationsOfProgress,
            lastProgressGeneration=cache.lastProgressGeneration,
            progress_halloffame=progress_hof,
        )

        with checkpointFile.open("wb") as cp_file:
            pickle.dump(cp, cp_file)

        # Begin the generational process
        for gen in range(start_gen, ngen + 1):
            generation_start = time.time()
            # Vary the population
            offspring = tools.selTournamentDCD(population, len(population))
            offspring = [cache.toolbox.clone(ind) for ind in offspring]

            for ind1, ind2 in zip(offspring[::2], offspring[1::2]):
                if random.random() <= cxpb:
                    cache.toolbox.mate(ind1, ind2)

                cache.toolbox.mutate(ind1)
                cache.toolbox.mutate(ind2)
                del ind1.fitness.values, ind2.fitness.values

            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
            stalled, stallWarn, progressWarn = util.eval_population(
                cache.toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, progress_hof, gen, result_data
            )

            gradCheck, newChildren = cache.toolbox.grad_search(gradCheck, offspring, cache, writer, csvfile, grad_hof, meta_hof, gen)
            offspring.extend(newChildren)

            progress.writeProgress(
                cache, gen, offspring, halloffame, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data
            )
            util.graph_process(cache, gen)
            util.graph_corner_process(cache, last=False)

            # Select the next generation population
            population = cache.toolbox.select(population + offspring, populationSize)

            if stallWarn:
                maxPopulation = cache.settings["maxPopulation"] * len(cache.MIN_VALUE)
                newPopulationSize = int(populationSize * stallRate)
                newPopulationSize += -newPopulationSize % 4
                newPopulationSize = min(newPopulationSize, maxPopulation)
                newPopulationSize += -newPopulationSize % 4

                diffSize = newPopulationSize - populationSize
                newPopulation = cache.toolbox.randomPopulation(n=diffSize)

                invalid_ind = [ind for ind in newPopulation if not ind.fitness.valid]
                util.eval_population(
                    cache.toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, progress_hof, gen, result_data
                )

                # This is just to assign the crowding distance to the individuals
                # no actual selection is done
                newPopulation = cache.toolbox.select(newPopulation, len(newPopulation))

                population.extend(newPopulation)
                populationSize = newPopulationSize

            if progressWarn:
                minPopulation = cache.settings["minPopulation"] * len(cache.MIN_VALUE)
                newPopulationSize = int(populationSize * progressRate)
                newPopulationSize += -newPopulationSize % 4
                newPopulationSize = max(newPopulationSize, minPopulation)
                newPopulationSize += -newPopulationSize % 4

                diffSize = populationSize - newPopulationSize

                population = cache.toolbox.select(population, newPopulationSize)
                populationSize = newPopulationSize

            cp = dict(
                population=population,
                generation=gen,
                halloffame=halloffame,
                rndstate=random.getstate(),
                gradCheck=gradCheck,
                meta_halloffame=meta_hof,
                grad_halloffame=grad_hof,
                generationsOfProgress=cache.generationsOfProgress,
                lastProgressGeneration=cache.lastProgressGeneration,
                progress_halloffame=progress_hof,
            )

            hof = Path(cache.settings["resultsDirMisc"], "hof")
            with hof.open("wb") as data:
                numpy.savetxt(data, numpy.array(halloffame))
            with checkpointFile.open("wb") as cp_file:
                pickle.dump(cp, cp_file)

            best = meta_hof.getBestScores()

            if stop_iteration(best, stalled, cache):
                util.finish(cache)
                util.graph_corner_process(cache, last=True)
                return halloffame
                
        util.finish(cache)
        util.graph_corner_process(cache, last=True)
        return halloffame



import sys

import matplotlib
import matplotlib.style as mplstyle
mplstyle.use('fast')

matplotlib.use("Agg")
import matplotlib.pyplot as plt

from cadet import H5

from CADETMatch.cache import cache

from pathlib import Path
import numpy
from sklearn.preprocessing import MinMaxScaler



def main():
    cache.setup(sys.argv[1])

    mcmcDir = Path(cache.settings["resultsDirMCMC"])

    kde_settings = H5()
    kde_settings.filename = (mcmcDir / "kde_settings.h5").as_posix()
    kde_settings.load()

    store = kde_settings.root.store

    plt.figure(figsize=[10, 10])
    plt.scatter(store[:, 0], store[:, 1])
    plt.xlabel("bandwidth")
    plt.ylabel("cross_val_score")
    plt.yscale('log')
    plt.xscale('log')
    plt.savefig(str(mcmcDir / "log_bandwidth.png"), bbox_inches="tight")
    plt.close()

    plt.figure(figsize=[10, 10])
    plt.scatter(store[:, 0], 1-store[:, 1])
    plt.xlabel("bandwidth")
    plt.ylabel("1 - cross_val_score")
    plt.yscale('log')
    plt.xscale('log')
    plt.savefig(str(mcmcDir / "1-log_bandwidth.png"), bbox_inches="tight")
    plt.close()

    plt.figure(figsize=[10, 10])
    plt.scatter(store[:, 0], store[:, 1])
    plt.xlabel("bandwidth")
    plt.ylabel("cross_val_score")
    plt.savefig(str(mcmcDir / "bandwidth.png"), bbox_inches="tight")
    plt.close()

    dir_base = cache.settings.get("resultsDirBase")
    file = dir_base / "kde_data.h5"

    kde_data = H5()
    kde_data.filename = file.as_posix()
    kde_data.load()

    times = {}
    values = {}
    for key in kde_data.root.keys():
        if "_time" in key:
            times[key.replace("_time", "")] = kde_data.root[key]
        if "_unit" in key:
            values[key] = kde_data.root[key]

    error_model = mcmcDir / "error_model"
    error_model.mkdir(parents=True, exist_ok=True)

    scaler = MinMaxScaler()
    prob = numpy.exp(kde_settings.root.probability)
    prob = prob[:kde_settings.root.scores.shape[0]]
    prob = numpy.squeeze(scaler.fit_transform(prob.reshape(-1, 1)))

    sort_index = numpy.argsort(prob)

    colors = plt.cm.rainbow(prob)

    for key, value in values.items():
        time = times[key.split("_unit", 1)[0]]
        plt.figure(figsize=[20, 10])
        for idx in sort_index:
            plt.plot(time, value[idx,:], color=colors[idx])
        plt.xlabel("time")
        plt.ylabel("concentration")
        sm = plt.cm.ScalarMappable(cmap=plt.cm.rainbow, norm=plt.Normalize(vmin=0, vmax=1))
        plt.colorbar(sm)
        plt.savefig((error_model / ("%s.png" % key)).as_posix(), bbox_inches="tight", dpi=600)
        plt.close()

    for exp_name in kde_data.root.errors:
        temp = kde_data.root.errors[exp_name].pump_delays.reshape(-1, 1)
        if numpy.any(temp > 0):
            plt.figure(figsize=[10,10])
            plt.hist(temp, bins=40)
            plt.savefig((error_model / ("%s_pump_delays.png" % exp_name)).as_posix(), bbox_inches="tight")
            plt.close()

        temp = kde_data.root.errors[exp_name].flow_rates.reshape(-1, 1)
        if numpy.any(temp > 0):
            plt.figure(figsize=[10,10])
            plt.hist(temp, bins=40)
            plt.savefig((error_model / ("%s_flow_rates.png" % exp_name)).as_posix(), bbox_inches="tight")
            plt.close()

        temp = kde_data.root.errors[exp_name].loading_concentrations.reshape(-1, 1)
        if numpy.any(temp > 0):
            plt.figure(figsize=[10,10])
            plt.hist(temp, bins=40)
            plt.savefig((error_model / ("%s_loading_concentrations.png" % exp_name)).as_posix(), bbox_inches="tight")
            plt.close()

    (error_model / "scores").mkdir(parents=True, exist_ok=True)
    for idx in range(kde_settings.root.scores.shape[1]):
        plt.figure(figsize=[10,10])
        temp = kde_settings.root.scores[:,idx]
        plt.hist(temp[temp > 0], bins=40)
        plt.savefig((error_model / "scores"/ ("%s.png" % idx)).as_posix(), bbox_inches="tight")
        plt.close()

    (error_model / "scores_mirror_scaled").mkdir(parents=True, exist_ok=True)
    for idx in range(kde_settings.root.scores_mirror_scaled.shape[1]):
        plt.figure(figsize=[10,10])
        plt.hist(kde_settings.root.scores_mirror_scaled[:,idx], bins=40)
        plt.savefig((error_model / "scores_mirror_scaled"/ ("%s.png" % idx)).as_posix(), bbox_inches="tight")
        plt.close()

if __name__ == "__main__":
    main()



import CADETMatch.util as util
from CADETMatch.abstract.transform import AbstractTransform


class SumTransform(AbstractTransform):
    @property
    def name(self):
        return "sum"

    @property
    def count(self):
        return 0

    @property
    def count_extended(self):
        return 0

    def transform(self):
        return []

    grad_transform = transform

    def untransform_inputorder(self, seq):
        return []

    def untransform(self, seq):
        return [], []

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        return None

    untransform_matrix_inputorder = untransform_matrix

    def setSimulation(self, sim, seq, experiment):
        if self.parameter.get("experiments", None) is None or experiment["name"] in self.parameter["experiments"]:
            location1 = self.parameter["location1"]
            location2 = self.parameter["location2"]
            locationSum = self.parameter["locationSum"]

            try:
                comp1 = self.parameter["component1"]
                bound1 = self.parameter["bound1"]
                index1 = None
            except KeyError:
                index1 = self.parameter["index1"]
                bound1 = None
                comp1 = None
            value1 = getValue(sim, location1, bound=bound1, comp=comp1, index=index1)

            try:
                comp2 = self.parameter["component2"]
                bound2 = self.parameter["bound2"]
                index2 = None
            except KeyError:
                index2 = self.parameter["index2"]
                bound2 = None
                comp2 = None

            value2 = getValue(sim, location2, bound=bound2, comp=comp2, index=index2)

            try:
                compSum = self.parameter["componentSum"]
                boundSum = self.parameter["boundSum"]
                indexSum = None
            except KeyError:
                indexSum = self.parameter["indexSum"]
                boundSum = None
                compSum = None
            setValue(sim, value1 + value2, locationSum, bound=boundSum, comp=compSum, index=indexSum)

        return [], []

    def getBounds(self):
        return None, None

    def getGradBounds(self):
        return None, None

    def getHeaders(self):
        return []

    def getHeadersActual(self):
        return self.getHeaders()

    def setBounds(self, parameter, lb, ub):
        return None


plugins = {"sum": SumTransform}



import CADETMatch.util as util
import CADETMatch.progress as progress
import CADETMatch.pareto as pareto
import csv
from pathlib import Path
import time
import CADETMatch.jacobian as jacobian
import multiprocessing
import array

name = "ScoreTest"


def run(cache, tools, creator):
    "run the parameter estimation"
    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    with path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)
        pop = cache.toolbox.population(n=0)
        sim_start = generation_start = time.time()
        result_data = {
            "input": [],
            "output": [],
            "output_meta": [],
            "results": {},
            "times": {},
            "input_transform": [],
            "input_transform_extended": [],
            "strategy": [],
            "mean": [],
            "confidence": [],
        }

        if "seeds" in cache.settings:
            seed_pop = [
                cache.toolbox.individual_guess([f(v) for f, v in zip(cache.settings["transform"], sublist)])
                for sublist in cache.settings["seeds"]
            ]
            pop.extend(seed_pop)

        if cache.metaResultsOnly:
            hof = pareto.DummyFront()
        else:
            hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
        meta_hof = pareto.ParetoFrontMeta(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))
        grad_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
        progress_hof = pareto.ParetoFrontMeta(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))

        invalid_ind = [ind for ind in pop if not ind.fitness.valid]
        stalled, stallWarn, progressWarn = util.eval_population(
            cache.toolbox, cache, invalid_ind, writer, csvfile, hof, meta_hof, progress_hof, -1, result_data
        )
        progress.writeProgress(cache, -1, pop, hof, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data)

        if cache.settings.get("condTest", None):
            for ind in invalid_ind:
                J = jacobian.jac(ind, cache)
                multiprocessing.get_logger().info("%s %s", ind, J)

        population = [cache.toolbox.individual_guess(i) for i in meta_hof]
        stalled, stallWarn, progressWarn = util.eval_population_final(
            cache.toolbox, cache, population, writer, csvfile, hof, meta_hof, progress_hof, 0, result_data
        )
        progress.writeProgress(cache, 0, population, hof, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data)

        util.finish(cache)
        util.graph_corner_process(cache, last=True)
        return hof


def setupDEAP(cache, fitness, fitness_final, grad_fitness, grad_search, grad_search_fine, map_function, creator, base, tools):
    "setup the DEAP variables"
    creator.create("FitnessMax", base.Fitness, weights=[1.0] * cache.numGoals)
    creator.create("Individual", list, typecode="d", fitness=creator.FitnessMax, strategy=None, mean=None, confidence=None, csv_line=None)

    creator.create("FitnessMaxMeta", base.Fitness, weights=[1.0, 1.0, 1.0, -1.0, -1.0])
    creator.create("IndividualMeta", array.array, typecode="d", fitness=creator.FitnessMaxMeta, strategy=None, csv_line=None, best=None)
    cache.toolbox.register("individualMeta", util.initIndividual, creator.IndividualMeta, cache)

    cache.toolbox.register(
        "individual", util.generateIndividual, creator.Individual, len(cache.MIN_VALUE), cache.MIN_VALUE, cache.MAX_VALUE, cache
    )
    cache.toolbox.register("population", tools.initRepeat, list, cache.toolbox.individual)

    cache.toolbox.register("individual_guess", util.initIndividual, creator.Individual, cache)

    cache.toolbox.register("evaluate", fitness, json_path=cache.json_path)
    cache.toolbox.register("evaluate_final", fitness_final, json_path=cache.json_path)
    cache.toolbox.register("evaluate_grad", grad_fitness, json_path=cache.json_path)
    cache.toolbox.register("evaluate_grad_fine", grad_search_fine, json_path=cache.json_path)
    cache.toolbox.register("grad_search", grad_search)

    cache.toolbox.register("map", map_function)



import CADETMatch.util as util
import CADETMatch.score as score
import numpy
import pandas
import scipy.optimize
from addict import Dict
import multiprocessing

name = "fractionationSlide"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 0
    settings.meta_mask = True

    data = pandas.read_csv(feature["fraction_csv"])
    headers = data.columns.values.tolist()
    comps = len(headers[2:])

    settings.count = 3 * comps

    return settings


def goal(offset, frac_exp, sim_data_time, spline, start, stop):
    frac_sim = util.fractionate_spline(start + offset, stop + offset, spline)
    return float(numpy.sum((frac_exp - frac_sim) ** 2))


def run(sim_data, feature):
    simulation = sim_data["simulation"]
    timeFunc = feature["timeFunc"]
    components = feature["components"]
    numComponents = len(components)
    samplesPerComponent = feature["samplesPerComponent"]
    data = feature["data"]
    CV_time = feature["CV_time"]
    start = feature["start"]
    stop = feature["stop"]
    funcs = feature["funcs"]

    time_center = (start + stop) / 2.0

    times = simulation.root.output.solution.solution_times

    scores = []

    sim_values_sse = []
    exp_values_sse = []

    graph_sim = {}
    graph_exp = {}
    graph_sim_offset = {}

    for component, value_func in funcs:
        exp_values = numpy.array(data[str(component)])
        selected = numpy.isfinite(exp_values)
        sim_value = simulation.root.output.solution[feature["unit"]]["solution_outlet_comp_%03d" % component]

        spline = scipy.interpolate.InterpolatedUnivariateSpline(times, sim_value, ext=1)

        lb = times[numpy.argmax(sim_value)] - times[-1]
        ub = times[numpy.argmax(sim_value)] - times[0]

        #getting a starting point estimate
        offsets = numpy.linspace(lb, ub, 100)
        errors = numpy.array([goal(offset, exp_values[selected], times, spline, start[selected], stop[selected]) for offset in offsets])
        idx_min = numpy.argmin(errors)

        offset_start, min_offsets, min_errors = util.find_opt_poly(offsets, errors, idx_min)

        result_powell = scipy.optimize.minimize(
            goal, offset_start, args=(exp_values[selected], times, spline, start[selected], stop[selected]), method="powell",
            bounds=[(min_offsets[0], min_offsets[-1]),]
        )

        time_offset = result_powell.x[0]

        fracOffset = util.fractionate_spline(start[selected] - time_offset, stop[selected] - time_offset, spline)

        # if the simulation scale and exp scale are too different the estimation of similarity, offset etc is not accurate discard if value max/min > 1e3
        max_exp = max(exp_values[selected])
        max_sim = max(fracOffset)
        if max(max_exp, max_sim) / min(max_exp, max_sim) > 1e3:
            value_score = 0
            pear = 0
            time_score = 0
        else:
            value_score = value_func(max(fracOffset))
            pear = score.pear_corr(scipy.stats.pearsonr(exp_values[selected], fracOffset)[0])
            time_score = timeFunc(abs(time_offset))

        exp_values_sse.extend(exp_values[selected])
        sim_values_sse.extend(fracOffset)

        scores.append(pear)
        scores.append(time_score)
        scores.append(value_score)

        graph_sim[component] = list(zip(start[selected], stop[selected], fracOffset))
        graph_exp[component] = list(zip(start[selected], stop[selected], exp_values[selected]))

        graph_sim_offset[component] = time_offset

    sim_data["graph_exp"] = graph_exp
    sim_data["graph_sim"] = graph_sim
    sim_data["graph_sim_offset"] = graph_sim_offset

    return (
        scores,
        util.sse(numpy.array(sim_values_sse), numpy.array(exp_values_sse)),
        len(sim_values_sse),
        time_center,
        numpy.array(sim_values_sse),
        numpy.array(exp_values_sse),
        [1.0 - i for i in scores],
    )


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    data = pandas.read_csv(feature["fraction_csv"])
    rows, cols = data.shape

    headers = data.columns.values.tolist()

    start = numpy.array(data.iloc[:, 0])
    stop = numpy.array(data.iloc[:, 1])

    smallestTime = min(start - stop)
    abstolFraction = abstol * smallestTime

    funcs = []

    for idx, component in enumerate(headers[2:], 2):
        value = numpy.array(data.iloc[:, idx])
        funcs.append((int(component), score.value_function(max(value), abstolFraction)))

    temp["data"] = data
    temp["start"] = start
    temp["stop"] = stop
    temp["timeFunc"] = score.time_function_decay(feature["time"][-1])
    temp["components"] = [int(i) for i in headers[2:]]
    temp["samplesPerComponent"] = rows
    temp["CV_time"] = CV_time
    temp["funcs"] = funcs
    temp["unit"] = feature["unit_name"]
    temp["peak_max"] = data.iloc[:, 2:].max().min()
    return temp


def headers(experimentName, feature):
    data = pandas.read_csv(feature["fraction_csv"])
    rows, cols = data.shape

    data_headers = data.columns.values.tolist()

    temp = []
    for component in data_headers[2:]:
        temp.append("%s_%s_Component_%s_Similarity" % (experimentName, feature["name"], component))
        temp.append("%s_%s_Component_%s_Time" % (experimentName, feature["name"], component))
        temp.append("%s_%s_Component_%s_Value" % (experimentName, feature["name"], component))
    return temp



import CADETMatch.util as util
import math
from CADETMatch.abstract.transform import AbstractTransform


class DiameterTransform(AbstractTransform):
    @property
    def name(self):
        return "null"

    @property
    def count(self):
        return 1

    @property
    def count_extended(self):
        return self.count

    def transform(self):
        def trans(i):
            return i

        return [
            trans,
        ]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        diameter = seq[0]
        return [diameter,]

    def untransform(self, seq):
        diameter = self.untransform_inputorder(seq)
        values = [
            math.pi * diameter ** 2 / 4.0,
        ]
        headerValues = [values[0], diameter]
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = numpy.zeros(matrix.shape)
        values[:, 0] = math.pi * matrix[:, 0] ** 2 / 4.0
        return values

    untransform_matrix_inputorder = untransform_matrix

    def setSimulation(self, sim, seq, experiment):
        values, headerValues = self.untransform(seq)

        if self.parameter.get("experiments", None) is None or experiment["name"] in self.parameter["experiments"]:
            location = self.parameter["location"]
            sim[location.lower()] = values[0]
        return values, headerValues

    def getBounds(self):
        minValue = self.parameter["min"]
        maxValue = self.parameter["max"]

        return [minValue,], [maxValue,]

    def getGradBounds(self):
        return [self.parameter["min"],], [self.parameter["max"],]

    def getHeaders(self):
        headers = []
        headers.append("Area")
        headers.append("Diameter")
        return headers

    def getHeadersActual(self):
        headers = []
        headers.append("Area")
        return headers

    def setBounds(self, parameter, lb, ub):
        parameter["min"] = lb[0]
        parameter["max"] = ub[0]


class NormDiameterTransform(DiameterTransform):
    @property
    def name(self):
        return "norm_diameter"

    def transform(self):
        minValue = self.parameter["min"]
        maxValue = self.parameter["max"]

        def trans(i):
            return (i - minValue) / (maxValue - minValue)

        return [
            trans,
        ]

    grad_transform = transform

    def untransform(self, seq):
        minValue = self.parameter["min"]
        maxValue = self.parameter["max"]

        values_transform = [
            (maxValue - minValue) * seq[0] + minValue,
        ]

        values = [
            math.pi * values_transform[0] ** 2 / 4.0,
        ]
        headerValues = [values[0], values_transform[0]]
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = self.untransform_matrix_inputorder(matrix)
        values[:, 0] = math.pi * values[:, 0] ** 2 / 4.0
        return values

    def untransform_matrix_inputorder(self, matrix):
        minValue = self.parameter["min"]
        maxValue = self.parameter["max"]

        values = (maxValue - minValue) * matrix + minValue
        return values

    def getBounds(self):
        return [0.0,], [1.0,]

    def getGradBounds(self):
        return [self.parameter["min"],], [self.parameter["max"],]


plugins = {"norm_diameter": NormDiameterTransform, "diameter": DiameterTransform}



import CADETMatch.util as util
import numpy
from CADETMatch.abstract.transform import AbstractTransform


class NuSigmaTransform(AbstractTransform):
    @property
    def name(self):
        return "nu_sigma"

    @property
    def count(self):
        return 2

    @property
    def count_extended(self):
        return 3

    def transform(self):
        def trans_a(i):
            return i

        def trans_b(i):
            return i

        return [trans_a, trans_b]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        nu = seq[0]
        nu_sigma = seq[1]
        return [nu, nu_sigma]

    def untransform(self, seq):
        nu, nu_sigma  = self.untransform_inputorder(seq)
        sigma = nu_sigma - nu
        values = [nu, sigma]
        headerValues = [nu, sigma, nu_sigma]
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = self.untransform_matrix_inputorder(matrix)
        values[:, 1] = matrix[:, 1] - matrix[:, 0]
        return values

    def untransform_matrix_inputorder(self, matrix):
        values = numpy.zeros(matrix.shape)
        values[:, 1] = matrix[:, 1]
        values[:, 0] = matrix[:, 0]
        return values

    def setSimulation(self, sim, seq, experiment):
        values, headerValues = self.untransform(seq)

        if self.parameter.get("experiments", None) is None or experiment["name"] in self.parameter["experiments"]:
            nu_location = self.parameter["nu_location"]
            sigma_location = self.parameter["sigma_location"]

            comp = self.parameter["component"]
            bound = self.parameter["bound"]

            unit = self.getUnit(nu_location)
            boundOffset = util.getBoundOffset(sim.root.input.model[unit])

            position = boundOffset[comp] + bound
            sim[nu_location.lower()][position] = values[0]
            sim[sigma_location.lower()][position] = values[1]

        return values, headerValues

    def getBounds(self):
        minNu = self.parameter["minNu"]
        maxNu = self.parameter["maxNu"]
        minSigma = self.parameter["minSigma"]
        maxSigma = self.parameter["maxSigma"]

        minValues = [minNu, minNu + minSigma]
        maxValues = [maxNu, maxNu + maxSigma]

        return minValues, maxValues

    def getGradBounds(self):
        return [self.parameter["min"],], [self.parameter["max"],]

    def getHeaders(self):
        nu_location = self.parameter["nu_location"]
        sigma_location = self.parameter["sigma_location"]
        nameNu = nu_location.rsplit("/", 1)[-1]
        nameSigma = sigma_location.rsplit("/", 1)[-1]
        bound = self.parameter["bound"]
        comp = self.parameter["component"]

        headers = []
        headers.append("%s Comp:%s Bound:%s" % (nameNu, comp, bound))
        headers.append("%s Comp:%s Bound:%s" % (nameSigma, comp, bound))
        headers.append("%s+%s Comp:%s Bound:%s" % (nameNu, nameSigma, comp, bound))
        return headers

    def getHeadersActual(self):
        nu_location = self.parameter["nu_location"]
        sigma_location = self.parameter["sigma_location"]
        nameNu = nu_location.rsplit("/", 1)[-1]
        nameSigma = sigma_location.rsplit("/", 1)[-1]
        bound = self.parameter["bound"]
        comp = self.parameter["component"]

        headers = []
        headers.append("%s Comp:%s Bound:%s" % (nameNu, comp, bound))
        headers.append("%s+%s Comp:%s Bound:%s" % (nameNu, nameSigma, comp, bound))
        return headers

    def setBounds(self, parameter, lb, ub):
        parameter["minNu"] = lb[0]
        parameter["maxNu"] = ub[0]
        parameter["minSigma"] = lb[1]
        parameter["maxSigma"] = ub[1]


class NormNuSigmaTransform(NuSigmaTransform):
    @property
    def name(self):
        return "norm_nu_sigma"

    def transform(self):
        minNu = self.parameter["minNu"]
        maxNu = self.parameter["maxNu"]
        minSigma = self.parameter["minSigma"]
        maxSigma = self.parameter["maxSigma"]

        def trans_a(i):
            return (i - minNu) / (maxNu - minNu)

        def trans_b(i):
            return (i - minSigma) / (maxSigma - minSigma)

        return [trans_a, trans_b]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        minNu = self.parameter["minNu"]
        maxNu = self.parameter["maxNu"]
        minSigma = self.parameter["minSigma"]
        maxSigma = self.parameter["maxSigma"]

        minValues = numpy.array([minNu, minNu + minSigma])
        maxValues = numpy.array([maxNu, maxNu + maxSigma])

        values = numpy.array(seq)

        values = (maxValues - minValues) * values + minValues
        return values

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = self.untransform_matrix_inputorder(matrix)
        values[:, 1] = values[:, 1] - values[:, 0]
        return values

    def untransform_matrix_inputorder(self, matrix):
        minNu = self.parameter["minNu"]
        maxNu = self.parameter["maxNu"]
        minSigma = self.parameter["minSigma"]
        maxSigma = self.parameter["maxSigma"]

        minValues = numpy.array([minNu, minNu + minSigma])
        maxValues = numpy.array([maxNu, maxNu + maxSigma])

        temp = (maxValues - minValues) * matrix + minValues

        values = numpy.zeros(temp.shape)
        values[:, 0] = temp[:, 0]
        values[:, 1] = temp[:, 1]

        return values

    def getBounds(self):
        return [0.0, 0.0], [1.0, 1.0]

    def getGradBounds(self):
        return [self.parameter["min"],], [self.parameter["max"],]


plugins = {"nu_sigma": NuSigmaTransform, "norm_nu_sigma": NormNuSigmaTransform}



import sys
from CADETMatch.cache import cache
from pathlib import Path

import matplotlib
import matplotlib.style as mplstyle
mplstyle.use('fast')

matplotlib.use("Agg")

import matplotlib.pyplot as plt
import matplotlib.animation as animation


def main():
    cache.setup(sys.argv[1])
    cache.progress_path = Path(cache.settings["resultsDirBase"]) / "progress.csv"

    output_spear = cache.settings["resultsDirSpace"] / "spearman"

    pngs = list(output_spear.glob("*.png"))

    pngs_sorted = sorted(pngs, key=lambda file: file.stat().st_ctime)

    fig = plt.figure(figsize=(10, 10), dpi=100)

    ims = []
    for png in pngs_sorted:
        with png.open("rb") as image_file:
            data = plt.imread(image_file)
        ymax, xmax, _ = data.shape
        im = plt.imshow(data, animated=True, aspect="equal", extent=[0, xmax, ymax, 0])
        plt.axis("off")
        ims.append([im])

    ani = animation.ArtistAnimation(fig, ims, interval=200, blit=False, repeat=False)

    FFMpegWriter = animation.writers["ffmpeg"]
    writer = FFMpegWriter(fps=5, bitrate=50000)
    fig.tight_layout()

    ani.save(str(output_spear / "spearman.mp4"), writer=writer)


if __name__ == "__main__":
    main()



import scipy.stats
import numpy
import scipy.optimize
import scipy.interpolate
import scipy.signal
import numpy.linalg
import CADETMatch.calc_coeff as calc_coeff
import CADETMatch.util as util
import multiprocessing
import sys
import SALib.sample.sobol_sequence
import math


def roll_spline(times, values, shift):
    "this function does approximately what the roll function does except that is used the spline so the shift does not have to be an integer and the points are resampled"

    spline = scipy.interpolate.InterpolatedUnivariateSpline(times, values, ext=3)

    times_new = times - shift

    values_new = spline(times_new)

    return values_new


def root_poly(corr, len_times, dt):
    index = numpy.argmax(corr)

    dt_pre = (index - len_times + 1) * dt

    if index == 0:
        indexes = numpy.array([index, index+1, index+2])
        multiprocessing.get_logger().warn("min index encountered in root poly  index: %s  dt: %s", index, dt_pre)
    elif index == (len(corr)-1):
        indexes = numpy.array([index-2, index-1, index])
        multiprocessing.get_logger().warn("max index encountered in root poly  index: %s  dt: %s", index, dt_pre)
    else:
        indexes = numpy.array([index-1, index, index+1])
        
    x = (indexes - len_times + 1) * dt
    y = corr[indexes]
    
    poly, res = numpy.polynomial.Polynomial.fit(x,y,2, full=True)
    
    dt = poly.deriv().roots()[0]
    return dt

def pearson_spline(exp_time_values, sim_data_values, exp_data_values):
    # resample to a much smaller time step to get a more precise offset
    sim_spline = scipy.interpolate.InterpolatedUnivariateSpline(exp_time_values, sim_data_values, ext=3)
    return pearson_spline_fun(exp_time_values, exp_data_values, sim_spline)

def pearson_spline_fun(exp_time_values, exp_data_values, sim_spline):
    # resample to a much smaller time step to get a more precise offset
    dt_approx = 1e-2
    points = math.ceil((exp_time_values[-1] - exp_time_values[0])/dt_approx)
    times = numpy.linspace(exp_time_values[0], exp_time_values[-1], points)
    dt = times[1] - times[0]

    sim_resample = sim_spline(times)
    exp_spline = scipy.interpolate.InterpolatedUnivariateSpline(exp_time_values, exp_data_values, ext=3)
    exp_resample = exp_spline(times)

    corr = scipy.signal.correlate(exp_resample, sim_resample)

    dt = root_poly(corr, len(times), dt)
    
    # calculate pearson correlation at the new time
    sim_data_values_copy = sim_spline(exp_time_values - dt)
    try:
        pear = scipy.stats.pearsonr(exp_data_values, sim_data_values_copy)[0]
    except ValueError:
        multiprocessing.get_logger().warn(
            "Pearson correlation failed to do NaN or InF in array  exp_array: [%s]   sim_array: [%s]",
            list(exp_data_values),
            list(sim_data_values_copy),
        )
        pear = 0
    score_local = pear_corr(pear)

    return score_local, dt


def time_function_decay(max_time):
    x_exp = numpy.array([0, max_time])
    y_exp = numpy.array([1, 0.0])

    a, b = calc_coeff.linear_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])

    def wrapper(diff):
        value = max(0.0, calc_coeff.linear(diff, a, b))

        return value

    return wrapper


def time_function(max_time, delay=10):
    x_lin_1 = numpy.array([0.0, delay])
    y_lin_1 = numpy.array([1.0, 0.95])

    x_lin_2 = numpy.array([delay, max_time])
    y_lin_2 = numpy.array([0.95, 0.0])

    a1, b1 = calc_coeff.linear_coeff(x_lin_1[0], y_lin_1[0], x_lin_1[1], y_lin_1[1])
    a2, b2 = calc_coeff.linear_coeff(x_lin_2[0], y_lin_2[0], x_lin_2[1], y_lin_2[1])

    def wrapper(diff):
        if diff <= delay:
            value = max(0.0, calc_coeff.linear(diff, a1, b1))
        else:
            value = max(0.0, calc_coeff.linear(diff, a2, b2))
        return value

    return wrapper


def value_function(peak_height, tolerance=1e-8, bottom_score=0.0):
    # if the peak height is 0 or less than the tolerance it needs to be treated as a special case to prevent divide by zero problems
    x = numpy.array([0.0, 1.0])
    y = numpy.array([1.0, bottom_score])

    # a, b = calc_coeff.exponential_coeff(x[0], y[0], x[1], y[1])
    a, b = calc_coeff.linear_coeff(x[0], y[0], x[1], y[1])

    if numpy.abs(peak_height) < tolerance:
        multiprocessing.get_logger().warn("peak height less than tolerance %s %s", tolerance, peak_height)

        def wrapper(x):
            if numpy.abs(x) < tolerance:
                return 1.0
            else:
                diff = numpy.abs(x - tolerance) / numpy.abs(tolerance)
                # return max(0, calc_coeff.exponential(diff, a, b))
                return max(0, calc_coeff.linear(diff, a, b))

    else:

        def wrapper(x):
            diff = numpy.abs(x - peak_height) / numpy.abs(peak_height)
            # return max(0, calc_coeff.exponential(diff, a, b))
            return max(0, calc_coeff.linear(diff, a, b))

    return wrapper


def value_function_exp(peak_height, tolerance=1e-8, bottom_score=0.05):
    # if the peak height is 0 or less than the tolerance it needs to be treated as a special case to prevent divide by zero problems
    x = numpy.array([0.0, 1.0])
    y = numpy.array([1.0, bottom_score])

    a, b = calc_coeff.exponential_coeff(x[0], y[0], x[1], y[1])
    # a, b = calc_coeff.linear_coeff(x[0], y[0], x[1], y[1])

    if numpy.abs(peak_height) < tolerance:
        multiprocessing.get_logger().warn("peak height less than tolerance %s %s", tolerance, peak_height)

        def wrapper(x):
            if numpy.abs(x) < tolerance:
                return 1.0
            else:
                diff = numpy.abs(x - tolerance) / numpy.abs(tolerance)
                return max(0, calc_coeff.exponential(diff, a, b))
                # return max(0, calc_coeff.linear(diff, a, b))

    else:

        def wrapper(x):
            diff = numpy.abs(x - peak_height) / numpy.abs(peak_height)
            return max(0, calc_coeff.exponential(diff, a, b))
            # return max(0, calc_coeff.linear(diff, a, b))

    return wrapper


def slope_function(peak_slope):
    # if the peak height is 0 or less than the tolerance it needs to be treated as a special case to prevent divide by zero problems
    x = numpy.array([0.0, 4.0])
    y = numpy.array([1.0, 0.0])

    a, b = calc_coeff.linear_coeff(x[0], y[0], x[1], y[1])

    def wrapper(x):
        diff = numpy.abs(x - peak_slope) / numpy.abs(peak_slope)
        return max(0, calc_coeff.linear(diff, a, b))

    return wrapper


def pear_corr(cr):
    # handle the case where a nan is returned
    if numpy.isnan(cr):
        return 0.0
    if cr < 0.0:
        return 0.0
    else:
        return cr

    # so far in looking cr has never been negative and the scores mostly just sit in the 0.8 to 0.99999 range
    # I am not even sure if cr could be negative with chromatography (indicating inverse relationship between simulation and experiment)


def cut_zero(times, values, min_value, max_value):
    "cut the raw times and values as close to min_value and max_value as possible and set the rest to zero without smoothing"
    data_zero = numpy.zeros(len(times))

    # find the starting point for the min_index and max_index, the real value could be off by 1 in either direction
    peak_max_index = numpy.argmax(values)

    if values[peak_max_index] < max_value:
        # the whole array is below the max_value we are looking for
        return values, None, None, None, None

    max_index = numpy.argmin((values[:peak_max_index] - max_value)**2)
    min_index = numpy.argmin((values[:max_index] - min_value)**2)

    min_time_found = times[min_index]
    min_value_found = values[min_index]

    max_time_found = times[max_index]
    max_value_found = values[max_index]

    data_zero[min_index : max_index + 1] = values[min_index : max_index + 1]

    return data_zero, min_time_found, min_value_found, max_time_found, max_value_found


def find_cuts(times, values, spline, spline_der):
    max_index = numpy.argmax(values)
    max_time = times[max_index]

    def goal(time):
        return -float(spline_der(time))

    time_space = numpy.linspace(0, max_time, int(max_time)*10)
    goal_space = -(spline_der(time_space))

    idx = numpy.argmin(goal_space)

    guess, time_found, goal_found = util.find_opt_poly(time_space, goal_space, idx)

    lb = time_found[0]
    ub = time_found[-1]

    result = scipy.optimize.minimize(goal, guess, method="powell", bounds=[(lb, ub),])

    max_time = float(result.x[0])
    max_value = float(spline(max_time))

    max_target_time = find_target(spline_der, 1, times, values)

    max_index = numpy.argmin((values - max_value)**2)

    min_time = find_target(spline, 1e-3 * max_value, times[:max_index], values[:max_index])
    min_value = float(spline(min_time))

    return min_time, min_value, max_time, max_value

def find_target(spline, target, times, values, rate=10):
    max_index = numpy.argmax(values)
    max_time = times[max_index]
    
    test_times = numpy.linspace(0, max_time, int(max_time)*rate)

    if not len(test_times):
        return None

    test_values = spline(test_times)
    
    error = (test_values - target)**2
    idx = numpy.argmin(error) 

    guess, time_found, goal_found = util.find_opt_poly(test_times, error, idx)

    lb = time_found[0]
    ub = time_found[-1]
    
    def goal(time):
        sse = float((spline(time) - target) ** 2)
        return sse

    result = scipy.optimize.minimize(goal, guess, method="powell", tol=1e-5, bounds=[(lb,ub),])

    found_time = float(result.x[0])
    found_value = spline(found_time)

    if result.success is False:
        multiprocessing.get_logger().info("target %s time %s value %s lb %s guess %s ub %s", target, found_time, 
                                          found_value, lb, guess, ub)
        return None    
    
    return found_time


import CADETMatch.util as util
import numpy
from CADETMatch.abstract.transform import AbstractTransform
import multiprocessing

class AutoKeqTransform(AbstractTransform):
    @property
    def name(self):
        return "auto_keq"

    @property
    def count(self):
        return 2

    @property
    def count_extended(self):
        return 3

    @property
    def okay_linear_ka(self):
        minValue = self.parameter["minKA"]
        maxValue = self.parameter["maxKA"]
        maxFactor = self.parameter.get('maxFactor', 1000)

        return maxValue/minValue < maxFactor

    @property
    def okay_log_ka(self):
        return not self.okay_linear_ka

    @property
    def okay_linear_keq(self):
        minValue = self.parameter["minKEQ"]
        maxValue = self.parameter["maxKEQ"]
        maxFactor = self.parameter.get('maxFactor', 1000)

        return maxValue/minValue < maxFactor

    @property
    def okay_log_keq(self):
        return not self.okay_linear_keq

    def transform(self):
        minKA = self.parameter["minKA"]
        maxKA = self.parameter["maxKA"]
        minKEQ = self.parameter["minKEQ"]
        maxKEQ = self.parameter["maxKEQ"]

        if self.okay_log_ka:
            def trans_a(i):
                return (numpy.log(i) - numpy.log(minKA)) / (numpy.log(maxKA) - numpy.log(minKA))
        else:
            def trans_a(i):
                return (i - minKA) / (maxKA - minKA)

        if self.okay_log_keq:
            def trans_b(i):
                return (numpy.log(i) - numpy.log(minKEQ)) / (numpy.log(maxKEQ) - numpy.log(minKEQ))
        else:
            def trans_b(i):
                return (i - minKEQ) / (maxKEQ - minKEQ)

        return [trans_a, trans_b]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        minKA = self.parameter["minKA"]
        maxKA = self.parameter["maxKA"]
        minKEQ = self.parameter["minKEQ"]
        maxKEQ = self.parameter["maxKEQ"]

        if self.okay_log_ka:
            ka = (numpy.log(maxKA) - numpy.log(minKA)) * seq[0] + numpy.log(minKA)
            ka = numpy.exp(ka)
        else:
            ka = (maxKA - minKA) * seq[0] + minKA

        if self.okay_log_keq:
            keq = (numpy.log(maxKEQ) - numpy.log(minKEQ)) * seq[1] + numpy.log(minKEQ)
            keq = numpy.exp(keq)
        else:
            keq = (maxKEQ - minKEQ) * seq[1] + minKEQ

        return [ka, keq]

    def untransform(self, seq):
        ka, keq = self.untransform_inputorder(seq)
        kd = ka/keq

        values = [ka, kd]
        headerValues = [ka, kd, keq]
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = self.untransform_matrix_inputorder(matrix)
        values[:, 1] = values[:, 0] / values[:, 1]
        return values

    def untransform_matrix_inputorder(self, matrix):
        minKA = self.parameter["minKA"]
        maxKA = self.parameter["maxKA"]
        minKEQ = self.parameter["minKEQ"]
        maxKEQ = self.parameter["maxKEQ"]

        if self.okay_log_ka:
            minKA = numpy.log(minKA)
            maxKA = numpy.log(maxKA)

        if self.okay_log_keq:
            minKEQ = numpy.log(minKEQ)
            maxKEQ = numpy.log(maxKEQ)

        minValues = numpy.array([minKA, minKEQ])
        maxValues = numpy.array([maxKA, maxKEQ])

        values = (maxValues - minValues) * matrix + minValues

        if self.okay_log_ka and self.okay_log_keq:
            values = numpy.exp(values)
        elif self.okay_log_ka and self.okay_linear_keq:
            values[:,0] = numpy.exp(values[:,0])
        elif self.okay_linear_ka and self.okay_log_keq:
            values[:,1] = numpy.exp(values[:,1])

        return values

    def setSimulation(self, sim, seq, experiment):
        values, headerValues = self.untransform(seq)

        if self.parameter.get("experiments", None) is None or experiment["name"] in self.parameter["experiments"]:
            location = self.parameter["location"]

            comp = self.parameter["component"]
            bound = self.parameter["bound"]

            unit = self.getUnit(location[0])
            boundOffset = util.getBoundOffset(sim.root.input.model[unit])

            position = boundOffset[comp] + bound
            sim[location[0].lower()][position] = values[0]
            sim[location[1].lower()][position] = values[1]

        return values, headerValues

    def getBounds(self):
        return [0.0, 0.0], [1.0, 1.0]

    def getGradBounds(self):
        return [self.parameter["min"],], [self.parameter["max"],]

    def getHeaders(self):
        if self.okay_log_ka:
            message_ka = "log ka"
        else:
            message_ka = "linear ka"

        if self.okay_log_keq:
            message_keq = "log keq"
        else:
            message_keq = "linear keq"

        multiprocessing.get_logger().info("parameter %s %s %s", self.parameter['location'], message_ka, message_keq)

        location = self.parameter["location"]
        nameKA = location[0].rsplit("/", 1)[-1]
        nameKD = location[1].rsplit("/", 1)[-1]
        bound = self.parameter["bound"]
        comp = self.parameter["component"]

        headers = []
        headers.append("%s Comp:%s Bound:%s" % (nameKA, comp, bound))
        headers.append("%s Comp:%s Bound:%s" % (nameKD, comp, bound))
        headers.append("%s/%s Comp:%s Bound:%s" % (nameKA, nameKD, comp, bound))
        return headers

    def getHeadersActual(self):
        location = self.parameter["location"]
        nameKA = location[0].rsplit("/", 1)[-1]
        nameKD = location[1].rsplit("/", 1)[-1]
        bound = self.parameter["bound"]
        comp = self.parameter["component"]

        headers = []
        headers.append("%s Comp:%s Bound:%s" % (nameKA, comp, bound))
        headers.append("%s/%s Comp:%s Bound:%s" % (nameKA, nameKD, comp, bound))
        return headers

    def setBounds(self, parameter, lb, ub):
        parameter["minKA"] = lb[0]
        parameter["maxKA"] = ub[0]
        parameter["minKEQ"] = lb[1]
        parameter["maxKEQ"] = ub[1]


plugins = {"auto_keq": AutoKeqTransform}




import CADETMatch.util as util
import CADETMatch.calc_coeff as calc_coeff
from addict import Dict

name = "AbsoluteHeight"

"""This score is NOT for optimization. It is needed for the MCMC algorithm in order to handle assymetric distributions"""


def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 0
    settings.meta_mask = False
    settings.count = 1
    return settings


def run(sim_data, feature):
    "similarity, value, start stop"
    selected = feature["selected"]
    exp_data_values = feature["value"][selected]
    sim_time_values, sim_data_values = util.get_times_values(sim_data["simulation"], feature)

    ys = [0.0, 1.0]
    xs = [0.0, max(exp_data_values) * 0.9]

    a, b = calc_coeff.linear_coeff(xs[0], ys[0], xs[1], ys[1])

    value = calc_coeff.linear(max(sim_data_values) * 0.9, a, b)

    temp = [
        value,
    ]

    return (
        temp,
        util.sse(sim_data_values, exp_data_values),
        len(sim_data_values),
        sim_time_values,
        sim_data_values,
        exp_data_values,
        [1.0 - i for i in temp],
    )


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    return {}


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    temp = ["%s_Height" % name]
    return temp



import sys

import matplotlib
import matplotlib.style as mplstyle
mplstyle.use('fast')

matplotlib.use("Agg")

size = 20

matplotlib.rc("font", size=size)  # controls default text sizes
matplotlib.rc("axes", titlesize=size)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=size)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=size)  # legend fontsize
matplotlib.rc("figure", titlesize=size)  # fontsize of the figure title
matplotlib.rc("figure", autolayout=True)

from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from mpl_toolkits.mplot3d import Axes3D

from CADETMatch.cache import cache

from pathlib import Path
import numpy
import scipy.interpolate
import itertools

from cadet import Cadet, H5
from addict import Dict

# parallelization
import multiprocessing

import os
import warnings
import CADETMatch.util as util
import logging
import CADETMatch.loggerwriter as loggerwriter

from emcee import autocorr

def main(map_function):
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "autocorr.log")
    cache.setup(sys.argv[1])

    multiprocessing.get_logger().info("autocorr graphing directory %s", os.getcwd())

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    mcmc_h5 = mcmcDir / "mcmc.h5"
    if mcmc_h5.exists():
        mcmc_store = H5()
        mcmc_store.filename = mcmc_h5.as_posix()
        mcmc_store.load(paths=["/full_chain", "/train_full_chain", "/bounds_full_chain"])

        progress_path = Path(cache.settings["resultsDirBase"]) / "result.h5"

        graph_dir = cache.settings["resultsDirSpace"] / "mcmc"

        input_headers = cache.parameter_headers_actual

        for chain in ("full_chain", "train_full_chain", "bounds_full_chain"):
            if chain in mcmc_store.root:
                plot_chain(input_headers, mcmc_store.root[chain], chain, graph_dir / chain)


def plot_chain(headers, chain, chain_name, graph_dir):
    graph_dir.mkdir(parents=True, exist_ok=True)

    for i in range(chain.shape[2]):

        fig = figure.Figure(figsize=[15, 7])
        canvas = FigureCanvas(fig)
        graph = fig.add_subplot(1, 1, 1)

        lines = []
        for j in range(chain.shape[0]):
            lines.append(autocorr.function_1d(chain[j, :, i]))
            graph.plot(lines[-1])
        graph.plot(numpy.mean(numpy.array(lines), axis=0), "k", linewidth=4)

        graph.set_xlabel("time")
        graph.set_ylabel("correlation")
        filename = "correlation_%s.png" % (headers[i])
        filename = filename.replace(":", "_").replace("/", "_")

        sum_lines = numpy.sum(numpy.array(lines), axis=0)
        sum_lines = sum_lines / chain.shape[0]
        taus = 2.0 * numpy.cumsum(sum_lines) - 1.0
        tau = taus[autocorr.auto_window(taus, 5)]

        graph.set_title("Correlation graph %s  Tau: %.2f" % (headers[i], tau))
        fig.savefig((graph_dir / filename).as_posix())

        fig = figure.Figure(figsize=[15, 7])
        canvas = FigureCanvas(fig)
        graph = fig.add_subplot(1, 1, 1)
        graph.set_title("Tau graph %s  Tau: %.2f" % (headers[i], tau))
        graph.set_xlabel("time")
        graph.set_ylabel("tau")
        graph.plot(taus, "k", linewidth=4)
        filename = "tau_%s.png" % (headers[i])
        filename = filename.replace(":", "_").replace("/", "_")
        fig.savefig((graph_dir / filename).as_posix())


if __name__ == "__main__":
    map_function = util.getMapFunction()
    main(map_function)
    sys.exit()



import CADETMatch.util as util
import CADETMatch.progress as progress
import CADETMatch.pareto as pareto
import csv
from pathlib import Path
from cadet import H5
import time
import CADETMatch.jacobian as jacobian
import multiprocessing
import array
import numpy

name = "AltScore"


def run(cache, tools, creator):
    "run the parameter estimation"
    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    with path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)
        pop = cache.toolbox.population(n=0)
        sim_start = generation_start = time.time()
        result_data = {
            "input": [],
            "output": [],
            "output_meta": [],
            "results": {},
            "times": {},
            "input_transform": [],
            "input_transform_extended": [],
            "strategy": [],
            "mean": [],
            "confidence": [],
        }

        previousResults = cache.settings["PreviousResults"]
        data = H5()
        data.filename = previousResults
        data.load(paths=["/meta_population","/meta_population_transform"])

        meta_population = data.root.meta_population
        meta_population_transform = data.root.meta_population_transform

        for line in meta_population:
            pop.append(cache.toolbox.individual_guess(line))

        multiprocessing.get_logger().info("altScore starting population %s %s", len(pop), pop)
        multiprocessing.get_logger().info("altScore starting population (transform) %s %s", len(meta_population_transform), meta_population_transform)

        convert = util.convert_population_inputorder(numpy.array(pop), cache)

        multiprocessing.get_logger().info("altScore starting population (convert) %s %s", len(convert), convert)

        if cache.metaResultsOnly:
            hof = pareto.DummyFront()
        else:
            hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
        meta_hof = pareto.ParetoFrontMeta(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))
        grad_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
        progress_hof = pareto.ParetoFrontMeta(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))

        invalid_ind = [ind for ind in pop if not ind.fitness.valid]
        stalled, stallWarn, progressWarn = util.eval_population(
            cache.toolbox, cache, invalid_ind, writer, csvfile, hof, meta_hof, progress_hof, -1, result_data
        )
        progress.writeProgress(cache, -1, pop, hof, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data)

        if cache.settings.get("condTest", None):
            for ind in invalid_ind:
                J = jacobian.jac(ind, cache)
                multiprocessing.get_logger().info("%s %s", ind, J)

        population = [cache.toolbox.individual_guess(i) for i in meta_hof]
        stalled, stallWarn, progressWarn = util.eval_population_final(
            cache.toolbox, cache, population, writer, csvfile, hof, meta_hof, progress_hof, 0, result_data
        )
        progress.writeProgress(cache, 0, population, hof, meta_hof, grad_hof, progress_hof, sim_start, generation_start, result_data)

        util.finish(cache)
        util.graph_corner_process(cache, last=True)
        return hof


def setupDEAP(cache, fitness, fitness_final, grad_fitness, grad_search, grad_search_fine, map_function, creator, base, tools):
    "setup the DEAP variables"
    creator.create("FitnessMax", base.Fitness, weights=[1.0] * cache.numGoals)
    creator.create("Individual", list, typecode="d", fitness=creator.FitnessMax, strategy=None, mean=None, confidence=None, csv_line=None)

    creator.create("FitnessMaxMeta", base.Fitness, weights=[1.0, 1.0, 1.0, -1.0, -1.0])
    creator.create("IndividualMeta", array.array, typecode="d", fitness=creator.FitnessMaxMeta, strategy=None, csv_line=None, best=None)
    cache.toolbox.register("individualMeta", util.initIndividual, creator.IndividualMeta, cache)

    cache.toolbox.register(
        "individual", util.generateIndividual, creator.Individual, len(cache.MIN_VALUE), cache.MIN_VALUE, cache.MAX_VALUE, cache
    )
    cache.toolbox.register("population", tools.initRepeat, list, cache.toolbox.individual)

    cache.toolbox.register("individual_guess", util.initIndividual, creator.Individual, cache)

    cache.toolbox.register("evaluate", fitness, json_path=cache.json_path)
    cache.toolbox.register("evaluate_final", fitness_final, json_path=cache.json_path)
    cache.toolbox.register("evaluate_grad", grad_fitness, json_path=cache.json_path)
    cache.toolbox.register("evaluate_grad_fine", grad_search_fine, json_path=cache.json_path)
    cache.toolbox.register("grad_search", grad_search)

    cache.toolbox.register("map", map_function)

